# coding: utf-8

"""
law config parser implementation.
"""

__all__ = [  # noqa
    "Config", "sections", "options", "keys", "items", "update", "include", "get", "getint",
    "getfloat", "getboolean", "get_default", "get_expanded", "get_expanded_int",
    "get_expanded_float", "get_expanded_bool", "get_expanded_boolean", "is_missing_or_none",
    "find_option", "add_section", "has_section", "remove_section", "set", "has_option",
    "remove_option",
]


import os
import re
import glob
import tempfile

import luigi
import six
from six.moves.configparser import ConfigParser

from law.util import no_value, brace_expand, str_to_int, merge_dicts, is_lazy_iterable


this_dir = os.path.dirname(os.path.abspath(__file__))

_set = set


def law_home_path(*paths):
    home = os.getenv("LAW_HOME") or os.path.expandvars(os.path.expanduser("$HOME/.law"))
    return os.path.normpath(os.path.join(home, *map(str, paths)))


class Config(ConfigParser):
    """
    Custom law configuration parser with a few additions on top of the standard python
    ``ConfigParser``. Most notably, this class adds config *inheritance* via :py:meth:`update` and
    :py:meth:`include`, a mechanism to synchronize with the luigi configuration parser, option
    referencing, and environment variable expansion.

    When *config_file* is set, it is loaded during setup. When empty, and *skip_fallbacks* is
    *False*, the default config file locations defined in :py:attr:`_config_files` are checked. By
    default, the default configuration :py:attr:`_default_config` is loaded, which can be prevented
    by setting *skip_defaults* to *True*.

    .. py:classattribute:: _instance

        type: :py:class:`Config`

        Global instance of this class.

    .. py:classattribute:: _default_config

        type: dict

        Default configuration.

    .. py:classattribute:: _config_files

        type: list

        List of configuration files that are checked during setup (unless *skip_fallbacks* is
        *True*). When a file exists, the check is stopped. Therefore, the order is important here.
    """

    _instance = None

    class Deferred(object):
        """
        Wrapper around callables representing deferred options.
        """

        str_repr = str(object())

        def __init__(self, func):
            self.func = func

        def __call__(self, *args, **kwargs):
            return self.func(*args, **kwargs)

        def __str__(self):
            # same string repr for all instances to identify them as deferred objects
            return self.str_repr

    # default config values, extended by those of contrib packages below this class
    _default_config = {
        "core": {
            "law_home": law_home_path(),
            "index_file": os.getenv("LAW_INDEX_FILE") or law_home_path("index"),
            "software_dir": os.getenv("LAW_SOFTWARE_DIR") or law_home_path("software"),
            "inherit": None,
            "extend": None,
            "sync_env": True,
            "sync_luigi_config": True,
        },
        "logging": {
            "law": os.getenv("LAW_LOG_LEVEL") or "WARNING",
        },
        "modules": {},
        "task": {
            "colored_repr": False,
            "colored_str": True,
            "interactive_format": "fancy",
            "interactive_line_breaks": True,
            "interactive_line_width": 0,
            "interactive_status_skip_seen": False,
        },
        "target": {
            "colored_repr": False,
            "colored_str": True,
            "expand_path_repr": False,
            "filesize_repr": False,
            "tmp_dir": os.getenv("LAW_TARGET_TMP_DIR") or tempfile.gettempdir(),
            "tmp_dir_perm": 0o0770,
            "default_local_fs": "local_fs",
        },
        "local_fs": {
            "base": "/",
            # defined by FileSystem
            "has_permissions": True,
            "default_file_perm": None,
            "default_dir_perm": None,
            "create_file_dir": True,
            # defined by LocalFileSystem
            "local_root_depth": 1,
        },
        "job": {
            "job_file_dir": os.getenv("LAW_JOB_FILE_DIR") or tempfile.gettempdir(),
            "job_file_dir_mkdtemp": True,
            "job_file_dir_cleanup": False,
        },
        "notifications": {
            "mail_recipient": None,
            "mail_sender": None,
            "mail_smtp_host": "127.0.0.1",
            "mail_smtp_port": 25,
        },
        "bash_sandbox": {
            "stagein_dir_name": "stagein",
            "stageout_dir_name": "stageout",
            "law_executable": "law",
            "login": False,
        },
        "bash_sandbox_env": {},
        "venv_sandbox": {
            "stagein_dir_name": "stagein",
            "stageout_dir_name": "stageout",
            "law_executable": "law",
        },
        "venv_sandbox_env": {},
    }

    _config_files = ["$LAW_CONFIG_FILE", "law.cfg", law_home_path("config"), "etc/law/config"]

    _option_ref_regex = re.compile(r"^\&(::(?P<section>[^\:]+))?::(?P<option>.+)$")

    _env_option_regex = re.compile(r"^LAW__([a-zA-Z0-9_]+)__([a-zA-Z0-9_]+)$")

    @classmethod
    def instance(cls, *args, **kwargs):
        """
        Creates an instance of this class with all *args* and *kwargs*, saves it in
        :py:attr:`_instance`, and returns it. When :py:attr:`_instance` was already set before, no
        new instance is created.
        """
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    @classmethod
    def _parse_option_ref(cls, value, default_section=None):
        m = cls._option_ref_regex.match(value)
        if not m:
            return None
        return (m.group("section") or default_section, m.group("option"))

    @classmethod
    def _expand_path(cls, path, expand_vars=True, expand_user=True):
        if expand_vars:
            ph = "__law_tilde__"
            path = path.replace(r"\~", ph)
            path = os.path.expanduser(path)
            path = path.replace(ph, "~")

        if expand_user:
            ph = "__law_dollar__"
            path = path.replace(r"\$", ph)
            path = os.path.expandvars(path)
            path = path.replace(ph, "$")

        return path

    def __init__(self, config_file="", skip_defaults=False, skip_fallbacks=False,
            skip_includes=False, skip_env_sync=False, skip_luigi_sync=False,
            skip_resolve_deferred=False):
        ConfigParser.__init__(self, allow_no_value=True)

        # lookup to correct config file
        self.config_file = None
        config_files = []
        if config_file:
            config_files.append(config_file)
        if not skip_fallbacks:
            config_files += self._config_files
        for cf in config_files:
            cf = os.path.expandvars(os.path.expanduser(str(cf)))
            cf = os.path.normpath(os.path.abspath(cf))
            if os.path.isfile(cf):
                self.config_file = cf
                break

        # helper to include additional configs
        def include_configs(filenames):
            if isinstance(filenames, six.string_types):
                filenames = [f.strip() for f in brace_expand(filenames.strip(), split_csv=True)]
            for filename in filenames or []:
                if not filename:
                    continue
                # try to resolve filename relative to the main config file
                if self.config_file:
                    basedir = os.path.dirname(self.config_file)
                    filename = os.path.normpath(os.path.join(basedir, filename))
                self.include(filename)

        # load defaults
        if not skip_defaults:
            self.update(self._default_config)

        # load the content of inherited configs
        if not skip_includes and self.config_file:
            # eagerly read the config file to get a glimpse of the files to inherit from
            c = self.__class__(self.config_file, skip_defaults=True, skip_fallbacks=True,
                skip_includes=True, skip_env_sync=True, skip_luigi_sync=True,
                skip_resolve_deferred=True)
            opt = "inherit"
            if c.has_option("core", "inherit_configs") and not c.get_expanded("core", "inherit"):
                # print a warning, not using the logger yet since it's not initialized at this point
                opt = "inherit_configs"
                print(
                    "INFO: the 'core.inherit_configs' option is deprecated and will be removed in a "
                    "future release of law; please use 'core.inherit' instead",
                )
            include_configs(c.get_expanded("core", opt, None))

        # load the actual config file if given
        if self.config_file:
            self.read(self.config_file)

        # load the content of extended configs
        if not skip_includes:
            opt = "extend"
            if self.has_option("core", "extend_configs") and not self.get_expanded("core", "extend"):
                # print a warning, not using the logger yet since it's not initialized at this point
                opt = "extend_configs"
                print(
                    "INFO: the 'core.extend_configs' option is deprecated and will be removed in a "
                    "future release of law; please use 'core.extend' instead",
                )
            include_configs(self.get_expanded("core", opt, None))

        # sync with environment variables
        if not skip_env_sync and self.get_expanded_bool("core", "sync_env"):
            self.sync_env()

        # sync with luigi configuration
        if not skip_luigi_sync and self.get_expanded_bool("core", "sync_luigi_config"):
            self.sync_luigi_config()

        # resolve deferred default values
        if not skip_resolve_deferred:
            self.resolve_deferred_defaults()

    def _convert_to_boolean(self, value):
        # py2 backport
        if six.PY3:
            return super(Config, self)._convert_to_boolean(value)

        if value.lower() not in self._boolean_states:
            raise ValueError("Not a boolean: {}".format(value))

        return self._boolean_states[value.lower()]

    def _get_type_converter(self, type, value):
        if type in (str, "str", "s"):
            return str
        if type in (int, "int", "i"):
            return str_to_int
        if type in (float, "float", "f"):
            return float
        if type in (bool, "bool", "boolean", "b"):
            if isinstance(value, six.string_types):
                return self._convert_to_boolean
            return bool

        raise ValueError("unknown 'type' argument ({}), must be 'str', 'int', 'float', or "
            "'bool'".format(type))

    def optionxform(self, option):
        """"""
        return option

    def options(self, section, prefix=None, expand_vars=True, expand_user=True):
        """
        Returns all options of a *section* in a list. When *prefix* is set, only options starting
        with that prefix are considered. Environment variable expansion is performed on every
        returned option name, depending on whether *expand_vars* and *expand_user* are *True*.
        """
        options = []
        for option in ConfigParser.options(self, section):
            if prefix and not option.startswith(prefix):
                continue
            option = self._expand_path(option, expand_vars=expand_vars, expand_user=expand_user)
            options.append(option)
        return options

    def keys(self, *args, **kwargs):
        # deprecation warning until v0.1 (also remove the entry in __all__ above)
        logger.warning("the use of {0}.keys() is deprecated, please use {0}.options() "
            "instead".format(self.__class__.__name__))

        return self.options(*args, **kwargs)

    def items(self, section, prefix=None, expand_vars=True, expand_user=True, **kwargs):
        """
        Returns a dictionary of key-value pairs for the given *section*. When *prefix* is set, only
        options starting with that prefix are considered. Environment variable expansion is
        performed on every returned option name and corresponding value, depending on whether
        *expand_vars* and *expand_user* are *True*. Internally, py:meth:`get_expanded` is used
        to perform value expansion and type interpolation, and is passed all *kwargs*.
        """
        options = self.options(section, prefix=prefix, expand_vars=expand_vars,
            expand_user=expand_user)
        return [
            (opt, self.get_expanded(section, opt, expand_vars=expand_vars,
                expand_user=expand_user, **kwargs))
            for opt in options
        ]

    def set(self, section, option, value=None):
        """
        Sets an *option* of an existing *section* to *value*. When *value* is *None*.
        """
        # serialize the value to a string representation
        if value is not None:
            if isinstance(value, (list, tuple, _set)) or is_lazy_iterable(value):
                value = ",".join(map(str, value))
            else:
                value = str(value)

        return ConfigParser.set(self, section, option, value)

    def update(self, data, overwrite=True, overwrite_sections=None, overwrite_options=None):
        """
        Updates the currently stored configuration with new *data*, given as a dictionary. When
        *overwrite_sections* is *False*, sections in *data* that are already present in the current
        config are skipped. When *overwrite_options* is *False*, existing options are not
        overwritten. When *None*, both *overwrite_sections* and *overwrite_options* default to
        *overwrite*.
        """
        if overwrite_sections is None:
            overwrite_sections = overwrite
        if overwrite_options is None:
            overwrite_options = overwrite

        for section, _data in six.iteritems(data):
            # add the section when it does not exist, and continue when it does but not overwriting
            if not self.has_section(section):
                self.add_section(section)
            elif not overwrite_sections:
                continue

            for option, value in six.iteritems(_data):
                # set the option when overwriting anyway, or when it does not exist
                if not self.has_option(section, option) or overwrite_options:
                    self.set(section, option, value)

    def include(self, filename, *args, **kwargs):
        """
        Updates the current config by that found in *filename*. All *args* and *kwargs* are
        forwarded to :py:meth:`update`.
        """
        p = self.__class__(filename, skip_defaults=True, skip_fallbacks=True, skip_env_sync=True,
            skip_luigi_sync=True)
        self.update(p._sections, *args, **kwargs)

    def get_default(self, section, option, default=no_value, type=None, expand_vars=False,
            expand_user=False, split_csv=False, dereference=True, default_when_none=True,
            _skip_refs=None):
        """ get_default(section, option, default=no_value, type=None, expand_vars=False, expand_user=False, split_csv=False, dereference=True, default_when_none=True)
        Returns the config value defined by *section* and *option*. When either the section or the
        option do not exist and a *default* value is provided, this value returned instead. When
        *type* is set, it must be either `"str"`, `"int"`, `"float"`, or `"boolean"`. When
        *expand_vars* is *True*, environment variables are expanded. When *expand_user* is *True*,
        user variables are expanded as well. Sequences of values can be identified, split by comma
        and returned as a list when *split_csv* is *True*, which will also trigger brace expansion.

        Also, options retrieved by this method are allowed to refer to values of other options
        within the config, even to those in other sections. The syntax for config references is
        ``&[::section]::option``. When no section is given, the value refers to an option in the
        same section. Example:

        .. code-block:: ini

            [my_section]
            a: 123
            b: &::a              # 123, refers to "a" in the same section

            [bar_section]
            a: &::my_section::a  # 123, refers to "a" in "my_section"

        This behavior is the default and, if desired, can be disabled by setting *dereference* to
        *False*. When the reference is not resolvable, the default value is returned.

        When *default_when_none* is *True*, a *default* value is provided, and the option was found
        but its value is *None* or ``"None"`` (case-insensitive), the *default* is returned.
        """  # noqa
        # return the default when either the section or the option does not exist
        default_set = default != no_value
        if (not self.has_section(section) or not self.has_option(section, option)) and default_set:
            return default

        # get the value
        value = self.get(section, option)

        # handle variable expansion and dereferencing when value is a string
        # (which should always be the case, but subclasses might overwrite get())
        if isinstance(value, six.string_types):
            # expand
            value = self._expand_path(value, expand_vars=expand_vars, expand_user=expand_user)

            # resolve references
            if dereference:
                ref = self._parse_option_ref(value, default_section=section)
                if ref:
                    # to avoid circular references, keep track of already resolved ones
                    if _skip_refs is None:
                        _skip_refs = []
                    elif ref in _skip_refs:
                        return default
                    _skip_refs.append(ref)

                    # return the referenced value
                    return self.get_default(*ref, default=default, type=type,
                        expand_vars=expand_vars, expand_user=expand_user, dereference=dereference,
                        default_when_none=default_when_none, _skip_refs=_skip_refs)

        # interpret None and "None" as missing?
        if default_when_none and default_set:
            if value is None:
                return default
            if isinstance(value, six.string_types) and value.lower() == "none":
                return default

        # helper for optional type conversion
        cast_type = lambda value: self._get_type_converter(type, value)(value) if type else value

        # do csv splitting if requested
        if split_csv:
            return [cast_type(v.strip()) for v in brace_expand(value, split_csv=True)]

        return cast_type(value)

    def get_expanded(self, *args, **kwargs):
        """
        Same as :py:meth:`get_default`, but *expandvars* and *expanduser* arguments are set to
        *True* by default.
        """
        kwargs.setdefault("expand_vars", True)
        kwargs.setdefault("expand_user", True)
        return self.get_default(*args, **kwargs)

    def get_expanded_int(self, *args, **kwargs):
        """
        Same as :py:meth:`get_expanded` with *type* set to ``int``.
        """
        kwargs["type"] = int
        return self.get_expanded(*args, **kwargs)

    def get_expanded_float(self, *args, **kwargs):
        """
        Same as :py:meth:`get_expanded` with *type* set to ``float``.
        """
        kwargs["type"] = float
        return self.get_expanded(*args, **kwargs)

    def get_expanded_bool(self, *args, **kwargs):
        """
        Same as :py:meth:`get_expanded` with *type* set to ``bool``.
        """
        kwargs["type"] = bool
        return self.get_expanded(*args, **kwargs)

    def get_expanded_boolean(self, *args, **kwargs):
        """
        Alias for :py:meth:`get_expanded_bool` for backwards compatibility.
        """
        return self.get_expanded_bool(*args, **kwargs)

    def is_missing_or_none(self, section, option):
        """
        Returns *True* if the value defined by *section* and *option* is missing or ``"None"``
        (case-insensitive), and *False* otherwise. Options without values and those pointing to
        unresolvable references are considered missing. Example:

        .. code-block:: ini

            [my_section]
            a: 123
            b: &::a
            c: &::not_there
            d: None
            e

        .. code-block:: python

            is_missing_or_none("my_section", "a")  # False
            is_missing_or_none("my_section", "b")  # False
            is_missing_or_none("my_section", "c")  # True
            is_missing_or_none("my_section", "d")  # True
            is_missing_or_none("my_section", "e")  # True
            is_missing_or_none("my_section", "f")  # True
        """
        value = self.get_expanded(section, option, default=no_value)
        if isinstance(value, six.string_types):
            value = value.lower()
        return value in ("none", None, no_value)

    def find_option(self, section, *options):
        """
        Returns the name of the first existing *option* for a given *section*.
        :py:meth:`is_missing_or_none` is used to check the existence. When none of the selected
        *options* exists, *None* is returned.
        """
        for option in options:
            if not self.is_missing_or_none(section, option):
                return option
        return None

    def sync_env(self):
        """
        Synchronizes options defined via environment variables in the format
        ``LAW__<section>__<option>``. The synchronization only works in case neither the section nor
        the option contain double underscores (which is anyway discouraged).
        """
        for name, value in os.environ.items():
            m = self._env_option_regex.match(name)
            if not m:
                continue

            section, option = m.groups()
            if not self.has_section(section):
                self.add_section(section)
            self.set(section, option, value)

    def sync_luigi_config(self, push=True, pull=True):
        """
        Synchronizes sections starting with ``"luigi_"`` with the luigi configuration parser. First,
        when *push* is *True*, (variable-expanded and dereferenced) options that exist in law but
        **not** in luigi are stored as defaults in the luigi config. Then, when *pull* is *True*,
        all luigi-related options in the law config are overwritten with those from luigi. This way,
        options set via luigi defaults (environment variables, global configuration files,
        `LUIGI_CONFIG_PATH`) always have precendence.
        """
        prefix = "luigi_"
        lparser = luigi.configuration.LuigiConfigParser.instance()

        if push:
            for section in self.sections():
                if not section.startswith(prefix):
                    continue
                lsection = section[len(prefix):]

                if not lparser.has_section(lsection):
                    lparser.add_section(lsection)

                for option in self.options(section):
                    if not lparser.has_option(lsection, option):
                        value = self.get_expanded(section, option)
                        lparser.set(lsection, option, value)

        if pull:
            for lsection in lparser.sections():
                section = prefix + lsection

                if not self.has_section(section):
                    self.add_section(section)

                for option, value in lparser.items(lsection):
                    self.set(section, option, value)

    def resolve_deferred_defaults(self):
        """
        Traverses all options, checks whether they are deferred callables and if so, resolves and
        sets them.
        """
        # TODO: priority based order?
        for section in self.sections():
            for option, value in self.items(section):
                if value == self.Deferred.str_repr:
                    value = self._default_config.get(section, {}).get(option, value)
                if isinstance(value, self.Deferred):
                    self.set(section, option, str(value(self)))


# add configs exposed by contrib packages to the default config
contrib_defaults = []
for contrib_init in glob.glob(os.path.join(this_dir, "contrib", "*", "__init__.py")):
    # get the path of the config file
    path = os.path.join(os.path.dirname(contrib_init), "config.py")
    if not os.path.exists(path):
        continue
    # load its content (not via import!)
    mod = {}
    with open(path, "r") as f:
        exec(f.read(), mod)
    defaults_func = mod.get("config_defaults")
    if not callable(defaults_func):
        raise AttributeError(
            "contrib config file {} does not contain callable 'config_defaults'".format(path),
        )
    defaults = defaults_func(Config._default_config)
    if not isinstance(defaults, dict):
        raise TypeError(
            "callable 'config_defaults' of {} did not return dictionary, but got {}".format(
                path, defaults,
            ),
        )
    contrib_defaults.append(defaults)

# merge
if contrib_defaults:
    merge_dicts(Config._default_config, *contrib_defaults, deep=True, inplace=True)


# register convenience functions on module-level
for name in __all__[__all__.index("sections"):]:
    def closure(name):
        config = Config.instance()
        func = getattr(config, name)

        def wrapper(*args, **kwargs):
            """
            Shorthand for :py:meth:`Config.{}` of the singleton instance :py:meth:`Config.instance`.
            """
            return func(*args, **kwargs)

        wrapper.__name__ = name
        wrapper.__doc__ = wrapper.__doc__.format(name)

        return wrapper

    locals()[name] = closure(name)


# trailing imports
from law.logger import get_logger

logger = get_logger(__name__)



# coding: utf-8

"""
Base definition of remote workflows based on job submission and status polling.
"""

__all__ = ["JobData", "BaseRemoteWorkflowProxy", "BaseRemoteWorkflow"]


import os
import sys
import time
import re
import copy
import random
import threading
from collections import OrderedDict, defaultdict
from abc import abstractmethod

import luigi
import six

from law.workflow.base import BaseWorkflow, BaseWorkflowProxy
from law.job.dashboard import NoJobDashboard
from law.parameter import NO_FLOAT, NO_INT, get_param, DurationParameter
from law.util import (
    no_value, is_number, colored, iter_chunks, merge_dicts, human_duration, DotDict, ShorthandDict,
    InsertableDict,
)
from law.logger import get_logger


logger = get_logger(__name__)


class JobData(ShorthandDict):
    """
    Sublcass of :py:class:`law.util.ShorthandDict` that adds shorthands for the attributes *jobs*,
    *unsubmitted_jobs*, *tasks_per_job*, and *dashboard_config*. This container object is used to
    store and keep track of per job information in :py:class:`BaseRemoteWorkflow`.

    .. py:classattribute:: dummy_job_id

        type: string

        A unique, dummy job id (``"dummy_job_id"``).
    """

    attributes = {
        "jobs": {},  # job_num -> job_data (see classmethod below)
        "unsubmitted_jobs": {},  # job_num -> branches
        "attempts": {},  # job_num -> current attempt
        "tasks_per_job": 1,
        "dashboard_config": {},
    }

    dummy_job_id = "dummy_job_id"

    @classmethod
    def job_data(cls, job_id=dummy_job_id, branches=None, status=None, code=None, error=None,
            extra=None, **kwargs):
        """
        Returns a dictionary containing default job submission information such as the *job_id*,
        task *branches* covered by the job, a job *status* string, a job return code, an *error*
        message, and *extra* data. Additional *kwargs* are accepted but _not_ stored.
        """
        return dict(job_id=job_id, branches=branches or [], status=status, code=code, error=error,
            extra=extra or {})

    def __len__(self):
        return len(self.jobs) + len(self.unsubmitted_jobs)

    def update(self, other):
        """"""
        other = dict(other)
        # ensure that keys (i.e. job nums) in job dicts are integers
        for key in ["jobs", "unsubmitted_jobs", "attempts"]:
            if key in other:
                cls = other[key].__class__
                other[key] = cls((int(job_num), val) for job_num, val in six.iteritems(other[key]))

        super(JobData, self).update(other)


class PollData(ShorthandDict):
    """
    Sublcass of :py:class:`law.util.ShorthandDict` that holds variable attributes used during job
    status polling: the maximum number of parallel running jobs, *n_parallel*, the minimum number of
    finished jobs to consider the task successful, *n_finished_min*, the maximum number of
    failed jobs to consider the task failed, *n_failed_max*, the number of currently active jobs,
    *n_active*.
    """

    attributes = {
        "n_parallel": None,
        "n_finished_min": None,
        "n_failed_max": None,
        "n_active": None,
    }


class BaseRemoteWorkflowProxy(BaseWorkflowProxy):
    """
    Workflow proxy base class for remote workflows.

    .. py:classattribute:: job_error_messages

        type: dict

        A dictionary containing short error messages mapped to job exit codes as defined in the
        remote job execution script.

    .. py:attribute:: show_errors

        type: int

        Numbers of errors to explicity show during job submission and status polling.

    .. py:attribute:: summarize_status_errors

        type: bool, int

        During status polling, when the number of errors exceeds :py:attr:`show_errors`, a summary
        of errors if shown when this flag is true. When a number is given, the summary is printed if
        the number of errors exceeds this value.

    .. py:attribute:: job_manager

        type: :py:class:`law.job.base.BaseJobManager'

        Reference to the job manager object that handles the actual job submission, status queries,
        etc. The instance is created and configured by :py:meth:`create_job_manager`.

    .. py:attribute:: job_file_factory

        type: :py:class:`law.job.base.BaseJobFileFactory`

        Reference to a job file factory. The instance is created and configured by
        :py:meth:`create_job_file_factory`.

    .. py:attribute:: job_data

        type: :py:class:`JobData`

        The job data object holding job submission and status information.

    .. py:attribute:: dashboard

        type: :py:class:`law.job.dashboard.BaseJobDashboard`

        Reference to the dashboard instance that is used by the workflow.

    .. py:attribute:: job_data_cls

        type: type (read-only)

        Class for instantiating :py:attr:`job_data`.
    """

    # job error messages for errors defined in the remote job script
    job_error_messages = {
        5: "input file rendering failed",
        10: "dashboard file failed",
        20: "bootstrap file failed",
        30: "bootstrap command failed",
        40: "law detection failed",
        50: "task dependency check failed",
        60: "task execution failed",
        70: "stageout file failed",
        80: "stageout command failed",
    }

    # configures how many job errors are fully shown
    show_errors = 5

    # control the printing of status error summaries
    summarize_status_errors = True

    # maximum number of parallel jobs
    n_parallel_max = sys.maxsize

    def __init__(self, *args, **kwargs):
        super(BaseRemoteWorkflowProxy, self).__init__(*args, **kwargs)

        task = self.task

        # the job submission file factory
        self.job_file_factory = None

        # the job dashboard
        self.dashboard = None

        # variable data that changes during / configures the job polling
        self.poll_data = PollData(
            n_parallel=None,
            n_finished_min=-1,
            n_failed_max=-1,
            n_active=0,
        )

        # job data
        self.job_data = self.job_data_cls(tasks_per_job=task.tasks_per_job)

        # setup the job mananger
        self.job_manager = self.create_job_manager(threads=task.submission_threads)
        self.job_manager.status_diff_styles["unsubmitted"] = ({"color": "green"}, {}, {})
        self._job_manager_setup_kwargs = no_value

        # boolean per job num denoting if a job should be / was skipped
        self._skip_jobs = {}

        # retry counts per job num
        self._job_retries = defaultdict(int)

        # flag that denotes whether a submission was done befire, set in run()
        self._submitted = False

        # set of existing branches that is kept track of during processing
        self._existing_branches = None

        # flag denoting if jobs were cancelled or cleaned up (i.e. controlled)
        self._controlled_jobs = False

        # tracking url
        self._tracking_url = None

        # flag denoting whether the first log file was already printed
        self._printed_first_log = False

        # lock to protect the dumping of submission data
        self._dump_lock = threading.Lock()

        # intially, set the number of parallel jobs which might change at some piont
        self._set_parallel_jobs(task.parallel_jobs)

        # cache of process resources per job number, set initially during the first call to
        # process_resources()
        self._initial_process_resources = None

    @property
    def job_data_cls(self):
        return JobData

    @abstractmethod
    def create_job_manager(self, **kwargs):
        """
        Hook to instantiate and return a class derived of :py:class:`law.job.base.BaseJobManager`.
        This method must be implemented by inheriting classes and should update and forward all
        *kwargs* to the constructor of the respective job manager.
        """
        return

    def setup_job_manager(self):
        """
        Hook invoked externally to further setup the job mananger or perform batch system related
        preparations, e.g. before jobs can be submitted. The returned keyword arguments will be
        forwarded to the submit, cancel, cleanup and query methods of the job mananger.
        """
        return {}

    def _setup_job_manager(self):
        if self._job_manager_setup_kwargs == no_value:
            self._job_manager_setup_kwargs = self.setup_job_manager()

        return self._job_manager_setup_kwargs or {}

    @abstractmethod
    def create_job_file_factory(self, **kwargs):
        """
        Hook to instantiate and return a class derived of
        :py:class:`law.job.base.BaseJobFileFactory`. This method must be implemented by inheriting
        classes and should update and forward all *kwargs* to the constructor of the respective job
        file factory.
        """
        return

    @abstractmethod
    def create_job_file(self, *args, **kwargs):
        """
        Creates a job file using the :py:attr:`job_file_factory`. The expected arguments depend on
        whether the job manager supports job grouping during submission
        (:py:attr:`BaseJobManager.job_grouping_submit`). If it does, two arguments containing the
        job number (*job_num*) and the list of branch numbers (*branches*) covered by the job. If
        job grouping is supported, a single dictionary mapping job numbers to covered branch values
        must be passed. In any case, the path(s) of job files are returned.

        This method must be implemented by inheriting classes.
        """
        return

    def destination_info(self):
        """
        Hook that can return a string containing information on the location that jobs are submitted
        to. The string is appended to submission and status messages.
        """
        return InsertableDict()

    def _destination_info_postfix(self):
        """
        Returns the destination info ready to be appended to a string.
        """
        dst_info = self.destination_info()
        if isinstance(dst_info, dict):
            dst_info = list(dst_info.values())
        if isinstance(dst_info, (list, tuple)):
            dst_info = ", ".join(map(str, dst_info))
        if dst_info != "":
            dst_info = ", {}".format(dst_info)
        return dst_info

    def get_extra_submission_data(self, job_file, job_id, config, log=None):
        """
        Hook that is called after job submission with the *job_file*, the returned *job_id*, the
        submission *config* and an optional *log* file to return extra data that is saved in the
        central job data.
        """
        extra = {}
        if log:
            extra["log"] = str(log)
        return extra

    @property
    def tracking_url(self):
        return self._tracking_url

    @tracking_url.setter
    def tracking_url(self, tracking_url):
        old_url = self.tracking_url
        self._tracking_url = tracking_url
        if tracking_url and self.task:
            self.task.set_tracking_url(tracking_url)
            if tracking_url != old_url:
                self.task.publish_message("tracking url: {}".format(tracking_url))

    @property
    def _cancel_jobs(self):
        """
        Property that is *True* when the :py:attr:`cancel_jobs` attribute exists and is *True*.
        """
        return isinstance(getattr(self.task, "cancel_jobs", None), bool) and self.task.cancel_jobs

    @property
    def _cleanup_jobs(self):
        """
        Property that is *True* when the :py:attr:`cleanup_jobs` attribute exists and is *True*.
        """
        return isinstance(getattr(self.task, "cleanup_jobs", None), bool) and self.task.cleanup_jobs

    def _get_existing_branches(self, sync=False, collection=None):
        if self._existing_branches is None:
            sync = True

        if sync:
            # initialize with set
            self._existing_branches = set()
            # add initial branches existing in output collection
            if collection is None:
                collection = self.get_cached_output().get("collection")
            if collection is not None:
                keys = collection.count(existing=True, keys=True)[1]
                self._existing_branches |= set(keys)

        return self._existing_branches

    def _can_skip_job(self, job_num, branches):
        """
        Returns *True* when a job can be potentially skipped, which is the case when all branch
        tasks given by *branches* are complete.
        """
        if job_num not in self._skip_jobs:
            existing_branches = self._get_existing_branches()
            self._skip_jobs[job_num] = all((b in existing_branches) for b in branches)

            # when the job is skipped, ensure that a job data entry exists and set the status
            if self._skip_jobs[job_num]:
                if job_num not in self.job_data.jobs:
                    self.job_data.jobs[job_num] = self.job_data_cls.job_data(branches=branches)
                if not self.job_data.jobs[job_num]["status"]:
                    self.job_data.jobs[job_num]["status"] = self.job_manager.FINISHED

        return self._skip_jobs[job_num]

    def _get_job_kwargs(self, name):
        attr = "{}_job_kwargs_{}".format(self.workflow_type, name)
        kwargs = getattr(self.task, attr, None)
        if kwargs is None:
            attr = "{}_job_kwargs".format(self.workflow_type)
            kwargs = getattr(self.task, attr)

        # when kwargs is not a dict, it is assumed to be a list whose
        # elements represent task attributes that are stored without the workflow type prefix
        if not isinstance(kwargs, dict):
            _kwargs = {}
            for param_name in kwargs:
                kwarg_name = param_name
                if param_name.startswith(self.workflow_type + "_"):
                    kwarg_name = param_name[len(self.workflow_type) + 1:]
                _kwargs[kwarg_name] = get_param(getattr(self.task, param_name))
            kwargs = _kwargs

        return kwargs

    def _set_parallel_jobs(self, n_parallel):
        if n_parallel <= 0:
            n_parallel = self.n_parallel_max

        # do nothing when the value does not differ from the current one
        if n_parallel == self.poll_data.n_parallel:
            return

        # add or remove the "unsubmitted" status from the job manager and adjust the last counts
        is_inf = n_parallel == self.n_parallel_max
        man = self.job_manager
        has_unsubmitted = man.status_names[0] == "unsubmitted"
        if is_inf:
            if has_unsubmitted:
                man.status_names.pop(0)
                man.last_counts.pop(0)
        else:
            if not has_unsubmitted:
                man.status_names.insert(0, "unsubmitted")
                man.last_counts.insert(0, 0)

        # set the value in poll_data
        self.poll_data.n_parallel = n_parallel

        return n_parallel

    def _status_error_pairs(self, job_num, job_data):
        return InsertableDict([
            ("job", job_num),
            ("branches", job_data["branches"]),
            ("id", job_data["job_id"]),
            ("status", job_data["status"]),
            ("code", job_data["code"]),
            ("error", job_data.get("error")),
            ("job script error", self.job_error_messages.get(job_data["code"], no_value)),
            ("log", job_data["extra"].get("log", no_value)),
        ])

    def _print_status_errors(self, failed_jobs):
        print("{} in task {}:".format(
            colored("{} failed job(s)".format(len(failed_jobs)), color="red", style="bright"),
            self.task.task_id,
        ))

        # prepare the decision for showing the error summary
        threshold = self.summarize_status_errors
        show_summary = threshold and (isinstance(threshold, bool) or len(failed_jobs) >= threshold)

        # show the first n errors
        for i, (job_num, data) in enumerate(six.iteritems(failed_jobs)):
            # get status pairs
            status_pairs = self._status_error_pairs(job_num, data)
            if isinstance(status_pairs, dict):
                status_pairs = list(status_pairs.items())

            # print the status line
            status_line = ", ".join([
                "{} {}".format(colored("{}:".format(key), style="bright"), value)
                for key, value in status_pairs
                if value != no_value
            ])
            print("    {}".format(status_line))

            if i >= self.show_errors and len(failed_jobs) > self.show_errors + 1:
                remaining = len(failed_jobs) - self.show_errors
                if remaining > 0:
                    print("    ... and {} more".format(remaining))
                break
        else:
            # all errors shown, no need for a summary
            show_summary = False

        # additional error summary
        if show_summary:
            # group by error code and status
            groups = OrderedDict()
            for job_num, data in six.iteritems(failed_jobs):
                key = (data["code"], data["status"])
                if key not in groups:
                    groups[key] = {"n_jobs": 0, "n_branches": 0, "log": None, "error": None}
                groups[key]["n_jobs"] += 1
                groups[key]["n_branches"] += len(data["branches"])
                if not groups[key]["error"]:
                    groups[key]["error"] = data["error"]
                if not groups[key]["log"]:
                    groups[key]["log"] = data["extra"].get("log")

            # show the summary
            print(colored("error summary:", color="red", style="bright"))
            for (code, status), stats in six.iteritems(groups):
                # status messsages of known error codes
                code_str = ""
                if code in self.job_error_messages:
                    code_str = " ({})".format(self.job_error_messages[code])
                # pairs for printing
                summary_pairs = [
                    ("status", status),
                    ("code", str(code) + code_str),
                    ("example error", stats.get("error")),
                ]
                # add an example log file
                if stats["log"]:
                    summary_pairs.append(("example log", stats["log"]))
                # print the line
                summary_line = ", ".join([
                    "{} {}".format(colored("{}:".format(key), style="bright"), value)
                    for key, value in summary_pairs
                ])
                print("    {n_jobs} jobs ({n_branches} branches) with {summary_line}".format(
                    summary_line=summary_line, **stats))

    @classmethod
    def merge_resources(cls, resources):
        merged = defaultdict(int)
        for res in (resources.values() if isinstance(resources, dict) else resources):
            for name, count in res.items():
                merged[name] += count
        return merged

    @classmethod
    def _maximum_resources(cls, resources, n_parallel):
        # cap by the maximum number of parallel jobs, but in a way such that maximizes the sum of
        # all possible resources that could be claimed at any given point in time
        # (for evenly distributed resources across jobs, this would not be necessary)
        if not resources:
            return {}

        if isinstance(resources, dict):
            resources = list(resources.values())

        # get all possible keys
        keys = set.union(*(set(r.keys()) for r in resources))

        # flatten and sort by increasing counts, then select the n_parallel last ones
        flat_resources = sorted([
            tuple(r.get(k, 0) for k in keys)
            for r in resources
        ])[-n_parallel:]

        # create sums across counts
        merged_counts = [
            sum(r[i] for r in flat_resources)
            for i in range(len(flat_resources[0]))
        ]

        return dict(zip(keys, merged_counts))

    def process_resources(self, force=False):
        task = self.task

        # collect resources over all branches if not just controlling running jobs
        resources = self._initial_process_resources
        if not task.is_controlling_remote_jobs() and (resources is None or force):
            get_job_resources = self._get_task_attribute("job_resources")
            branch_chunks = iter_chunks(task.branch_map.keys(), task.tasks_per_job)
            resources = {
                job_num: get_job_resources(job_num, branches)
                for job_num, branches in enumerate(branch_chunks, 1)
                if not self._can_skip_job(job_num, branches)
            }
            # store resources only if the branch map was frozen
            if task.cache_branch_map:
                self._initial_process_resources = resources

        if not resources:
            return {}

        return self._maximum_resources(resources, self.poll_data.n_parallel)

    def complete(self):
        if self.task.is_controlling_remote_jobs():
            return self._controlled_jobs

        return super(BaseRemoteWorkflowProxy, self).complete()

    def requires(self):
        # use upstream and workflow specific requirements only when not controlling running jobs
        if self.task.is_controlling_remote_jobs():
            reqs = DotDict()
        else:
            reqs = super(BaseRemoteWorkflowProxy, self).requires()
            remote_reqs = self._get_task_attribute("workflow_requires")()
            if remote_reqs:
                reqs.update(remote_reqs)

        return reqs

    def output(self):
        """
        Returns the default workflow outputs in an ordered dictionary. At the moment, this is the
        collection of outputs of the branch tasks (key ``"collection"``), the submission file (key
        ``"submission"``), and the status file (key ``"status"``). These two *control outputs* are
        optional, i.e., they are not considered when checking the task's completeness.
        """
        task = self.task

        # get the directory where the control outputs are stored
        out_dir = self._get_task_attribute("output_directory")()

        # define outputs
        outputs = DotDict()
        postfix = [
            task.control_output_postfix(),
            self._get_task_attribute("output_postfix")(),
        ]
        postfix = "_".join(map(str, filter(bool, postfix))) or ""
        if postfix:
            postfix = "_" + postfix

        # file containing the job data, i.e. job ids, last status codes, log files, etc
        job_data_file = "{}_jobs{}.json".format(self.workflow_type, postfix)
        outputs["jobs"] = out_dir.child(job_data_file, type="f", optional=True)

        # update with upstream output when not just controlling running jobs
        if not task.is_controlling_remote_jobs():
            outputs.update(super(BaseRemoteWorkflowProxy, self).output())

        return outputs

    def dump_job_data(self):
        """
        Dumps the current submission data to the submission file.
        """
        # renew the dashboard config
        self.job_data["dashboard_config"] = self.dashboard.get_persistent_config()

        # write the job data to the output file
        output = self.get_cached_output()
        with self._dump_lock:
            output["jobs"].dump(self.job_data, formatter="json", indent=4)

        logger.debug("job data dumped")

    def get_run_context(self):
        return self._get_task_attribute("workflow_run_context")()

    def run(self):
        with self.get_run_context():
            super(BaseRemoteWorkflowProxy, self).run()
            return self._run_impl()

    def _run_impl(self):
        """
        Actual run method that starts the processing of jobs and initiates the status polling, or
        performs job cancelling or cleaning, depending on the task parameters.
        """
        task = self.task
        output = self.get_cached_output()

        # create the job dashboard interface
        self.dashboard = task.create_job_dashboard() or NoJobDashboard()

        # read job data and reset some values
        self._submitted = not task.ignore_submission and output["jobs"].exists()
        if self._submitted:
            # load job data and cast job ids
            self.job_data.update(output["jobs"].load(formatter="json"))
            for job_data in six.itervalues(self.job_data.jobs):
                job_data["job_id"] = self.job_manager.cast_job_id(job_data["job_id"])

            # sync other settings
            task.tasks_per_job = self.job_data.tasks_per_job
            self.dashboard.apply_config(self.job_data.dashboard_config)

        # store initially complete branches
        outputs_existing = False
        if "collection" in output:
            collection = output["collection"]
            count = len(self._get_existing_branches(collection=collection))
            outputs_existing = count >= collection._abs_threshold()

        # cancel jobs?
        if self._cancel_jobs:
            if self._submitted:
                self.cancel()
            self._controlled_jobs = True
            return

        # cleanup jobs?
        if self._cleanup_jobs:
            if self._submitted:
                self.cleanup()
            self._controlled_jobs = True
            return

        # were all outputs already existing?
        if outputs_existing:
            return

        # from here on, submit and/or wait while polling

        # maybe set a tracking url
        self.tracking_url = self.dashboard.create_tracking_url()

        # ensure the output directory exists
        if not self._submitted:
            output["jobs"].parent.touch()

        try:
            # instantiate the configured job file factory
            self.job_file_factory = self.create_job_file_factory()

            # submit
            if not self._submitted:
                # set the initial list of unsubmitted jobs
                branches = sorted(task.branch_map.keys())
                branch_chunks = list(iter_chunks(branches, task.tasks_per_job))
                self.job_data.unsubmitted_jobs = OrderedDict(
                    (i + 1, branches) for i, branches in enumerate(branch_chunks)
                )
                self.submit()

            # sleep once to give the job interface time to register the jobs
            if not self._submitted and not task.no_poll:
                post_submit_delay = self._get_task_attribute("post_submit_delay")()
                if post_submit_delay > 0:
                    logger.debug("sleep for {} second(s) due to post_submit_delay".format(
                        post_submit_delay))
                    time.sleep(post_submit_delay)

            # start status polling when a) no_poll is not set, or b) the jobs were already
            # submitted so that failed jobs are resubmitted after a single polling iteration
            if not task.no_poll or self._submitted:
                self.poll()

        finally:
            # in any event, cleanup the job file
            if self.job_file_factory:
                self.job_file_factory.cleanup_dir(force=False)

    def cancel(self):
        """
        Cancels running jobs. The job ids are read from the submission file which has to exist
        for obvious reasons.
        """
        task = self.task

        # get job ids from submission data
        job_ids = [
            d["job_id"] for d in self.job_data.jobs.values()
            if d["job_id"] not in (self.job_data.dummy_job_id, None)
        ]
        if not job_ids:
            return

        # setup the job manager
        job_man_kwargs = self._setup_job_manager()

        # get job kwargs for cancelling
        cancel_kwargs = merge_dicts(job_man_kwargs, self._get_job_kwargs("cancel"))

        # cancel jobs
        task.publish_message("going to cancel {} jobs".format(len(job_ids)))
        if self.job_manager.job_grouping_cancel:
            errors = self.job_manager.cancel_group(job_ids, **cancel_kwargs)
        else:
            errors = self.job_manager.cancel_batch(job_ids, **cancel_kwargs)

        # print errors
        if errors:
            print("{} error(s) occured while cancelling {} job(s) of task {}:".format(
                len(errors), len(job_ids), task.task_id))
            tmpl = "    {}"
            for i, err in enumerate(errors):
                print(tmpl.format(err))
                if i + 1 >= self.show_errors:
                    remaining = len(errors) - self.show_errors
                    if remaining > 0:
                        print("    ... and {} more".format(remaining))
                    break

        # inform the dashboard
        for job_num, job_data in six.iteritems(self.job_data.jobs):
            task.forward_dashboard_event(self.dashboard, job_data, "action.cancel", job_num)

    def cleanup(self):
        """
        Cleans up jobs on the remote run location. The job ids are read from the submission file
        which has to exist for obvious reasons.
        """
        task = self.task

        # get job ids from submission data
        job_ids = [
            d["job_id"] for d in self.job_data.jobs.values()
            if d["job_id"] not in (self.job_data.dummy_job_id, None)
        ]
        if not job_ids:
            return

        # setup the job manager
        job_man_kwargs = self._setup_job_manager()

        # get job kwargs for cleanup
        cleanup_kwargs = merge_dicts(job_man_kwargs, self._get_job_kwargs("cleanup"))

        # cleanup jobs
        task.publish_message("going to cleanup {} jobs".format(len(job_ids)))
        if self.job_manager.job_grouping_cleanup:
            errors = self.job_manager.cleanup_group(job_ids, **cleanup_kwargs)
        else:
            errors = self.job_manager.cleanup_batch(job_ids, **cleanup_kwargs)

        # print errors
        if errors:
            print("{} error(s) occured while cleaning up {} job(s) of task {}:".format(
                len(errors), len(job_ids), task.task_id))
            tmpl = "    {}"
            for i, err in enumerate(errors):
                print(tmpl.format(err))
                if i + 1 >= self.show_errors:
                    remaining = len(errors) - self.show_errors
                    if remaining > 0:
                        print("    ... and {} more".format(remaining))
                    break

    def submit(self, retry_jobs=None):
        """
        Submits all jobs. When *retry_jobs* is *None*, a new job list is built. Otherwise,
        previously failed jobs defined in the *retry_jobs* dictionary, which maps job numbers to
        lists of branch numbers, are used.
        """
        task = self.task

        # collect data of jobs that should be submitted: num -> branches
        submit_jobs = OrderedDict()

        # keep track of the list of unsubmitted job nums before retry jobs are handled to control
        # whether they are resubmitted immediately or at the end (subject to shuffling)
        unsubmitted_job_nums = list(self.job_data.unsubmitted_jobs.keys())
        if task.shuffle_jobs:
            random.shuffle(unsubmitted_job_nums)

        # handle jobs for resubmission
        if retry_jobs:
            for job_num, branches in six.iteritems(retry_jobs):
                # some retry jobs might be skipped
                if self._can_skip_job(job_num, branches):
                    continue

                # in the case that the number of parallel jobs might be reached, or when retry jobs
                # are configured to be tried last, add the jobs back to the unsubmitted ones and
                # update the job id
                n = self.poll_data.n_active + len(submit_jobs)
                if n >= self.poll_data.n_parallel or task.append_retry_jobs:
                    self.job_data.jobs.pop(job_num, None)
                    self.job_data.unsubmitted_jobs[job_num] = branches
                    if task.append_retry_jobs:
                        unsubmitted_job_nums.append(job_num)
                    else:
                        unsubmitted_job_nums.insert(0, job_num)
                    continue

                # mark job for resubmission
                submit_jobs[job_num] = sorted(branches)

        # fill with unsubmitted jobs until maximum number of parallel jobs is reached
        for job_num in unsubmitted_job_nums:
            branches = self.job_data.unsubmitted_jobs[job_num]

            # remove jobs that don't need to be submitted
            if self._can_skip_job(job_num, branches):
                self.job_data.unsubmitted_jobs.pop(job_num, None)
                continue

            # mark job for submission only when n_parallel is not reached yet
            n = self.poll_data.n_active + len(submit_jobs)
            if n < self.poll_data.n_parallel:
                self.job_data.unsubmitted_jobs.pop(job_num, None)
                submit_jobs[job_num] = sorted(branches)

        # store submission data for jobs about to be submitted
        new_submission_data = OrderedDict()

        # when there is nothing to submit, dump the submission data to the output file and stop here
        if not submit_jobs:
            if retry_jobs or self.job_data.unsubmitted_jobs:
                self.dump_job_data()
            return new_submission_data

        # add empty job entries to submission data
        for job_num, branches in six.iteritems(submit_jobs):
            job_data = self.job_data_cls.job_data(branches=branches)
            self.job_data.jobs[job_num] = job_data

        # log some stats
        dst_info = self._destination_info_postfix()
        task.publish_message("going to submit {} {} job(s){}".format(
            len(submit_jobs), self.workflow_type, dst_info))

        # job file preparation and submission
        if self.job_manager.job_grouping_submit:
            job_ids, submission_data = self._submit_group(submit_jobs)
        else:
            job_ids, submission_data = self._submit_batch(submit_jobs)

        # store submission data
        errors = []
        for job_num, job_id, data in six.moves.zip(submit_jobs, job_ids, submission_data.values()):
            # handle errors
            if isinstance(job_id, Exception):
                errors.append((job_num, job_id))
                job_id = self.job_data.dummy_job_id

            # set the job id in the job data
            job_data = self.job_data.jobs[job_num]
            job_data["job_id"] = job_id
            extra = self.get_extra_submission_data(data["job"], job_id, data["config"],
                log=data.get("log"))
            job_data["extra"].update(extra)
            new_submission_data[job_num] = copy.deepcopy(job_data)

            # inform the dashboard
            task.forward_dashboard_event(self.dashboard, job_data, "action.submit", job_num)

        # dump the job data to the output file
        self.dump_job_data()

        # raise exceptions or log
        if errors:
            print("{} error(s) occured during job submission of task {}:".format(
                len(errors), task.task_id))
            tmpl = "    job {}: {}"
            for i, tpl in enumerate(errors):
                print(tmpl.format(*tpl))
                if i + 1 >= self.show_errors:
                    remaining = len(errors) - self.show_errors
                    if remaining > 0:
                        print("    ... and {} more".format(remaining))
                    break
        else:
            task.publish_message(
                "submitted {} {} job(s){}".format(len(submit_jobs), self.workflow_type, dst_info),
            )

        return new_submission_data

    def _submit_batch(self, submit_jobs, **kwargs):
        task = self.task

        # create job submission files mapped to job nums
        all_job_files = OrderedDict()
        for job_num, branches in six.iteritems(submit_jobs):
            all_job_files[job_num] = self.create_job_file(job_num, branches)

        # all_job_files is an ordered mapping job_num -> {"job": PATH, "log": PATH/None},
        # get keys and values for faster lookup by numeric index
        job_nums = list(all_job_files.keys())
        job_files = [f["job"] for f in six.itervalues(all_job_files)]

        # prepare objects for dumping intermediate job data
        dump_freq = self._get_task_attribute("dump_intermediate_job_data")()
        if dump_freq and not is_number(dump_freq):
            dump_freq = 50

        # setup the job manager
        job_man_kwargs = self._setup_job_manager()

        # get job kwargs for submission and merge with passed kwargs
        submit_kwargs = merge_dicts(job_man_kwargs, self._get_job_kwargs("submit"), kwargs)

        # progress callback to inform the scheduler
        def progress_callback(i, job_id):
            job_num = job_nums[i]

            # some job managers respond with a list of job ids per submission (e.g. htcondor, slurm)
            # so get the first id as long as batched submission is not yet supported
            if isinstance(job_id, list) and not self.job_manager.chunk_size_submit:
                job_id = job_id[0]

            # set the job id early
            self.job_data.jobs[job_num]["job_id"] = job_id

            # log a message every 25 jobs
            if i in (0, len(job_files) - 1) or (i + 1) % 25 == 0:
                task.publish_message("submitted {}/{} job(s)".format(i + 1, len(job_files)))

            # dump intermediate job data with a certain frequency
            if dump_freq and (i + 1) % dump_freq == 0:
                self.dump_job_data()

        # submit
        job_ids = self.job_manager.submit_batch(
            job_files,
            retries=3,
            threads=task.submission_threads,
            callback=progress_callback,
            **submit_kwargs  # noqa
        )

        return (
            job_ids,
            all_job_files,
        )

    def _submit_group(self, submit_jobs, **kwargs):
        task = self.task

        # create the single multi submission file, passing the job_num -> branches dict
        job_file = self.create_job_file(submit_jobs)

        # setup the job manager
        job_man_kwargs = self._setup_job_manager()

        # get job kwargs for submission and merge with passed kwargs
        submit_kwargs = merge_dicts(job_man_kwargs, self._get_job_kwargs("submit"), kwargs)

        # submission
        job_ids = self.job_manager.submit_group(
            [job_file["job"]] * len(submit_jobs),
            retries=3,
            threads=task.submission_threads,
            **submit_kwargs  # noqa
        )

        # set all job ids
        for job_num, job_id in zip(submit_jobs, job_ids):
            self.job_data.jobs[job_num]["job_id"] = job_id

        return (
            job_ids,
            {job_num: job_file for job_num in submit_jobs},
        )

    def poll(self):
        """
        Initiates the job status polling loop.
        """
        task = self.task
        dump_intermediate_job_data = self._get_task_attribute("dump_intermediate_job_data")()

        # total job count
        n_jobs = len(self.job_data)

        # track finished and failed jobs in dicts holding status data
        finished_jobs = set()
        failed_jobs = set()

        # the resources of yet unfinished jobs as claimed initially and reported to the scheduler
        # and the maximum amount resources potentially claimed by the jobs
        job_resources = dict(self._initial_process_resources or {})
        max_resources = self._maximum_resources(job_resources, self.poll_data.n_parallel)

        # track number of consecutive polling failures and the start time
        n_poll_fails = 0
        start_time = time.time()

        # setup the job manager
        job_man_kwargs = self._setup_job_manager()

        # get job kwargs for status querying
        query_kwargs = merge_dicts(job_man_kwargs, self._get_job_kwargs("query"))

        # start the poll loop
        i = -1
        while True:
            i += 1

            # sleep after the first iteration
            if i > 0:
                time.sleep(task.poll_interval * 60)

            # handle scheduler messages, which could change some task parameters
            task._handle_scheduler_messages()

            # walltime exceeded?
            if task.walltime != NO_FLOAT and (time.time() - start_time) > task.walltime * 3600:
                raise Exception("exceeded walltime: {}".format(human_duration(hours=task.walltime)))

            # update variable attributes for polling
            self.poll_data.n_finished_min = task.acceptance * (1 if task.acceptance > 1 else n_jobs)
            self.poll_data.n_failed_max = task.tolerance * (1 if task.tolerance > 1 else n_jobs)

            # determine the currently active jobs, i.e., the jobs whose states should be checked,
            # jobs whose ids are unknown, and jobs that can be skipped, i.e., jobs whose tasks'
            # outputs are already present
            active_jobs = []
            unknown_jobs = []
            for job_num, data in six.iteritems(self.job_data.jobs):
                # skip jobs that are already known to be finished or failed
                if job_num in finished_jobs or job_num in failed_jobs:
                    continue

                # skip jobs whose tasks are aready complete
                if self._can_skip_job(job_num, data["branches"]):
                    data["status"] = self.job_manager.FINISHED
                    data["code"] = 0
                    finished_jobs.add(job_num)
                    continue

                # mark as active or unknown
                if data["job_id"] in (None, self.job_data.dummy_job_id):
                    data["job_id"] = self.job_data.dummy_job_id
                    data["status"] = self.job_manager.RETRY
                    data["error"] = "unknown job id"
                    unknown_jobs.append(job_num)
                else:
                    active_jobs.append(job_num)
            self.poll_data.n_active = len(active_jobs) + len(unknown_jobs)

            # query job states
            job_ids = [self.job_data.jobs[job_num]["job_id"] for job_num in active_jobs]
            if self.job_manager.job_grouping_query:
                query_data = self.job_manager.query_group(job_ids, **query_kwargs)
            else:
                query_data = self.job_manager.query_batch(job_ids, **query_kwargs)

            # separate into actual states and errors that might have occured during the status query
            states_by_id = OrderedDict()
            errors = []
            for job_num, (job_id, state_or_error) in zip(active_jobs, six.iteritems(query_data)):
                if isinstance(state_or_error, Exception):
                    errors.append(state_or_error)
                    continue

                # set the data
                states_by_id[job_id] = state_or_error

                # sync extra info if available
                extra = state_or_error.get("extra")
                if not isinstance(extra, dict):
                    continue
                self.job_data.jobs[job_num]["extra"].update(extra)

                # set tracking url on the workflow proxy
                if not self.tracking_url and "tracking_url" in extra:
                    self.tracking_url = extra["tracking_url"]

                # print the first log file
                if not self._printed_first_log and extra.get("log"):
                    task.publish_message("first log file: {}".format(extra["log"]))
                    self._printed_first_log = True
            del query_data

            # print the first couple errors
            if errors:
                print("{} error(s) occured during job status query of task {}:".format(
                    len(errors), task.task_id))
                tmpl = "    {}"
                for i, err in enumerate(errors):
                    print(tmpl.format(err))
                    if i + 1 >= self.show_errors:
                        remaining = len(errors) - self.show_errors
                        if remaining > 0:
                            print("    ... and {} more".format(remaining))
                        break

                # increase the fail counter and maybe stop with an exception
                n_poll_fails += 1
                if task.poll_fails > 0 and n_poll_fails > task.poll_fails:
                    raise Exception("poll_fails exceeded")

                # poll again
                continue
            else:
                # no errors occured, reset the fail counter
                n_poll_fails = 0

            # handle active jobs
            for job_num in active_jobs:
                # update job data with status info
                data = self.job_data.jobs[job_num]
                for field in ["status", "error", "code"]:
                    value = states_by_id[data["job_id"]][field]
                    if value is not None:
                        data[field] = value

                # when the task picked up an existing submission file, then in the first polling
                # iteration it might happen that a job is finished, but outputs of its tasks are
                # not existing, e.g. when they were removed externally and the job id is still known
                # to the batch system; in this case, mark it as unknown and to be retried
                if self._submitted and i == 0:
                    is_finished = data["status"] == self.job_manager.FINISHED
                    if is_finished and not self._can_skip_job(job_num, data["branches"]):
                        data["status"] = self.job_manager.RETRY
                        data["error"] = "initially missing task outputs"
            del states_by_id

            # from here on, consider unknown jobs again as active
            active_jobs += unknown_jobs
            del unknown_jobs

            # get settings from the task for triggering post-finished status checks
            check_completeness = self._get_task_attribute("check_job_completeness")()
            check_completeness_delay = self._get_task_attribute("check_job_completeness_delay")()
            if check_completeness_delay:
                time.sleep(check_completeness_delay)

            # store jobs per status and take further actions depending on the status
            pending_jobs = set()
            running_jobs = set()
            retry_jobs = set()
            newly_failed_jobs = []  # need to preserve order
            for job_num in active_jobs:
                data = self.job_data.jobs[job_num]

                if data["status"] == self.job_manager.PENDING:
                    pending_jobs.add(job_num)
                    task.forward_dashboard_event(self.dashboard, copy.deepcopy(data),
                        "status.pending", job_num)
                    continue

                if data["status"] == self.job_manager.RUNNING:
                    running_jobs.add(job_num)
                    task.forward_dashboard_event(self.dashboard, copy.deepcopy(data),
                        "status.running", job_num)
                    continue

                if data["status"] == self.job_manager.FINISHED:
                    # additionally check if the outputs really exist
                    if not check_completeness or all(
                        self.task.as_branch(b).complete()
                        for b in data["branches"]
                    ):
                        finished_jobs.add(job_num)
                        self._existing_branches |= set(data["branches"])
                        self.poll_data.n_active -= 1
                        data["job_id"] = self.job_data.dummy_job_id
                        task.forward_dashboard_event(self.dashboard, copy.deepcopy(data),
                            "status.finished", job_num)
                        # potentially clear logs
                        if task.clear_logs:
                            log_file = data["extra"].get("log")
                            if isinstance(log_file, six.string_types) and os.path.exists(log_file):
                                try:
                                    os.remove(log_file)
                                except (FileNotFoundError if six.PY3 else IOError):
                                    pass
                        continue

                    # the job is marked as finished but not all branches are complete
                    data["status"] = self.job_manager.FAILED
                    data["error"] = "branch task(s) incomplete due to missing outputs"

                if data["status"] in (self.job_manager.FAILED, self.job_manager.RETRY):
                    newly_failed_jobs.append(job_num)
                    self.poll_data.n_active -= 1

                    # retry or ultimately failed?
                    if self._job_retries[job_num] < task.retries:
                        self._job_retries[job_num] += 1
                        self.job_data.attempts.setdefault(job_num, 0)
                        self.job_data.attempts[job_num] += 1
                        data["status"] = self.job_manager.RETRY
                        retry_jobs.add(job_num)
                        task.forward_dashboard_event(self.dashboard, copy.deepcopy(data),
                            "status.retry", job_num)
                    else:
                        failed_jobs.add(job_num)
                        task.forward_dashboard_event(self.dashboard, copy.deepcopy(data),
                            "status.failed", job_num)
                    continue

                raise Exception("unknown job status '{}'".format(data["status"]))

            # gather some counts
            n_pending = len(pending_jobs)
            n_running = len(running_jobs)
            n_finished = len(finished_jobs)
            n_retry = len(retry_jobs)
            n_failed = len(failed_jobs)
            n_unsubmitted = len(self.job_data.unsubmitted_jobs)

            # log the status line
            counts = (n_pending, n_running, n_finished, n_retry, n_failed)
            if self.poll_data.n_parallel != self.n_parallel_max:
                counts = (n_unsubmitted,) + counts
            status_line = self.job_manager.status_line(counts, last_counts=True, sum_counts=n_jobs,
                color=True, align=task.align_polling_status_line)
            status_line += self._destination_info_postfix()
            status_line = task.modify_polling_status_line(status_line)
            task.publish_message(status_line)
            self.last_status_counts = counts

            # inform the scheduler about the progress
            task.publish_progress(100.0 * n_finished / n_jobs)

            # remove resources of finished and failed jobs
            for job_num in finished_jobs | failed_jobs:
                job_resources.pop(job_num, None)
            # check if the maximum possible resources decreased and report to the scheduler
            new_max_resources = self._maximum_resources(job_resources, self.poll_data.n_parallel)
            decreased_resources = {}
            for key, value in max_resources.items():
                diff = value - new_max_resources.get(key, 0)
                if diff > 0:
                    decreased_resources[key] = diff
            if decreased_resources:
                task.decrease_running_resources(decreased_resources)
            # some resources might even have increased (due to post-facto, user induced changes)
            max_resources = new_max_resources

            # print newly failed jobs
            if newly_failed_jobs:
                self._print_status_errors({
                    job_num: self.job_data.jobs[job_num]
                    for job_num in newly_failed_jobs
                })

            # infer the overall status
            reached_end = n_jobs == n_finished + n_failed
            finished = n_finished >= self.poll_data.n_finished_min
            failed = n_failed > self.poll_data.n_failed_max
            unreachable = n_jobs - n_failed < self.poll_data.n_finished_min

            # write job data
            if finished or dump_intermediate_job_data:
                self.dump_job_data()

            # stop when finished
            if finished:
                break

            # complain when failed
            if failed:
                failed_nums = sorted(failed_jobs - retry_jobs)
                raise Exception(
                    "tolerance exceeded for job(s) {}".format(",".join(map(str, failed_nums))),
                )

            # stop early if unreachable
            if unreachable:
                err = None
                if reached_end:
                    err = "acceptance of {} not reached, total jobs: {}, failed jobs: {}"
                elif task.check_unreachable_acceptance:
                    err = "acceptance of {} unreachable, total jobs: {}, failed jobs: {}"
                if err:
                    raise Exception(err.format(self.poll_data.n_finished_min, n_jobs, n_failed))

            # invoke the poll callback
            poll_callback_res = self._get_task_attribute("poll_callback")(self.poll_data)
            if poll_callback_res is False:
                logger.debug(
                    "job polling loop gracefully stopped due to False returned by poll_callback",
                )
                break

            # trigger automatic resubmission and submission of unsubmitted jobs if necessary
            if retry_jobs or self.poll_data.n_active < self.poll_data.n_parallel:
                self.submit({
                    job_num: self.job_data.jobs[job_num]["branches"]
                    for job_num in newly_failed_jobs
                })

            # break when no polling is desired
            # we can get to this point when there was already a submission and the no_poll
            # parameter was set so that only failed jobs are resubmitted once
            if task.no_poll:
                break

        duration = round(time.time() - start_time)
        task.publish_message("polling took {}".format(human_duration(seconds=duration)))


class BaseRemoteWorkflow(BaseWorkflow):
    """
    Opinionated base class for remote workflows that works in 2 phases:

        1. Create and submit *m* jobs that process *n* tasks. Submission information (mostly job ids)
        is stored in the so-called *jobs* file, which is an output target of this workflow.

        2. Use the job data and start status polling. When done, status data is stored alongside the
        submission information in the same *jobs* file.

    .. py:classattribute:: check_unreachable_acceptance

        type: bool

        When *True*, stop the job status polling early if the minimum number of finsihed jobs as
        defined by :py:attr:`acceptance` becomes unreachable. Otherwise, keep polling until all jobs
        are either finished or failed. Defaults to *False*.

    .. py:classattribute:: align_polling_status_line

        type: int, bool

        Alignment value that is passed to :py:meth:`law.job.base.BaseJobManager.status_line` to
        print the status line during job status polling. Defaults to *False*.

    .. py:classattribute:: append_retry_jobs

        type: bool

        When *True*, jobs to retry are added to the end of the jobs to submit, giving priority to
        new ones. However, when *shuffle_jobs* is *True*, they might be submitted again earlier.
        Defaults to *False*.

    .. py:classattribute:: include_member_resources

        type: bool

        When *True*, the task resources defined in :py:meth:`process_resources` will contain the
        the ones defined in the :py:attr:`resources` instance or class attribute. Defaults to
        *False*.

    .. py:classattribute:: retries

        type: :py:class:`luigi.IntParameter`

        Maximum number of automatic resubmission attempts per job before considering it failed.
        Defaults to *5*.

    .. py:classattribute:: tasks_per_job

        type: :py:class:`luigi.IntParameter`

        Number of tasks to be processed by per job. Defaults to *1*.

    .. py:classattribute:: parallel_jobs

        type: :py:class:`luigi.IntParameter`

        Maximum number of parallel running jobs, e.g. to protect a very busy queue of a batch
        system. Empty default value (infinity).

    .. py:classattribute:: no_poll

        type: :py:class:`luigi.BoolParameter`

        When *True*, only submit jobs and skip status polling. Defaults to *False*.

    .. py:classattribute:: submission_threads

        type: :py:class:`luigi.IntParameter`

        Number of threads to use for both job submission and job status polling. Defaults to 4.

    .. py:classattribute:: walltime

        type: :py:class:`law.DurationParameter`

        Maximum job walltime after which a job will be considered failed. Empty default value. The
        default unit is hours when a plain number is passed.

    .. py:classattribute:: job_workers

        type: :py:class:`luigi.IntParameter`

        Number of cores to use within jobs to process multiple tasks in parallel (via adding
        '--workers' to remote job command). Defaults to 1.

    .. py:classattribute:: poll_interval

        type: :py:class:`law.DurationParameter`

        Interval between two job status polls. Defaults to 1 minute. The default unit is minutes
        when a plain number is passed.

    .. py:classattribute:: poll_fails

        type: :py:class:`luigi.IntParameter`

        Maximum number of consecutive errors during status polling after which a job is considered
        failed. This can occur due to networking problems. Defaults to *5*.

    .. py:classattribute:: shuffle_jobs

        type: :py:class:`luigi.BoolParameter`

        When *True*, the order of jobs is shuffled before submission. Defaults to *False*.

    .. py:classattribute:: cancel_jobs

        type: :py:class:`luigi.BoolParameter`

        When *True*, already running jobs are cancelled and no new ones are submitted. The job ids
        are read from the jobs file if existing. Defaults to *False*.

    .. py:classattribute:: cleanup_jobs

        type: :py:class:`luigi.BoolParameter`

        When *True*, already running jobs are cleaned up and no new ones are submitted. The job ids
        are read from the jobs file if existing. Defaults to *False*.

    .. py:classattribute:: transfer_logs

        type: :py:class:`luigi.BoolParameter`

        Transfer the combined log file back to the output directory. Defaults to *False*.
    """

    retries = luigi.IntParameter(
        default=5,
        significant=False,
        description="number of automatic resubmission attempts per job; default: 5",
    )
    tasks_per_job = luigi.IntParameter(
        default=1,
        significant=False,
        description="number of tasks to be processed by one job; default: 1",
    )
    parallel_jobs = luigi.IntParameter(
        default=NO_INT,
        significant=False,
        description="maximum number of parallel running jobs; default: infinite",
    )
    no_poll = luigi.BoolParameter(
        default=False,
        significant=False,
        description="just submit, do not initiate status polling after submission; default: False",
    )
    submission_threads = luigi.IntParameter(
        default=4,
        significant=False,
        description="number of threads to use for (re)submission and status queries; default: 4",
    )
    poll_interval = DurationParameter(
        default=1,
        unit="m",
        significant=False,
        description="time between status polls; default unit is minutes; default: 1",
    )
    poll_fails = luigi.IntParameter(
        default=5,
        significant=False,
        description="maximum number of consecutive errors during polling; default: 5",
    )
    walltime = DurationParameter(
        default=NO_FLOAT,
        unit="h",
        significant=False,
        description="maximum wall time; default unit is hours; default: infinite",
    )
    job_workers = luigi.IntParameter(
        default=1,
        significant=False,
        description="number of cores to use within jobs to process multiple tasks in parallel (via "
        "adding --workers to remote job commands); default: 1",
    )
    shuffle_jobs = luigi.BoolParameter(
        default=False,
        significant=False,
        description="shuffled job submission; default: False",
    )
    cancel_jobs = luigi.BoolParameter(
        default=False,
        description="cancel all submitted jobs but do not submit new jobs; default: False",
    )
    cleanup_jobs = luigi.BoolParameter(
        default=False,
        description="cleanup all submitted jobs but do not submit new jobs; default: False",
    )
    ignore_submission = luigi.BoolParameter(
        default=False,
        significant=False,
        description="ignore any existing submission file from a previous submission and start a "
        "new one; default: False",
    )
    transfer_logs = luigi.BoolParameter(
        default=False,
        significant=False,
        description="transfer job logs to the output directory; default: False",
    )
    clear_logs = luigi.BoolParameter(
        default=False,
        significant=False,
        description="when --transfer-logs is set, remove logs of successful jobs; default: False",
    )

    check_unreachable_acceptance = False
    align_polling_status_line = False
    append_retry_jobs = False
    include_member_resources = False

    exclude_index = True

    exclude_params_branch = {
        "retries", "tasks_per_job", "parallel_jobs", "no_poll", "submission_threads", "walltime",
        "job_workers", "poll_interval", "poll_fails", "shuffle_jobs", "cancel_jobs", "cleanup_jobs",
        "ignore_submission", "transfer_logs",
    }
    exclude_params_repr = {"cancel_jobs", "cleanup_jobs"}

    exclude_params_remote_workflow = set()

    # for internal consistency, inheriting classes should implement
    #  - <workflow_type>_job_resources
    #  - <workflow_type>_workflow_requires
    #  - <workflow_type>_output_directory
    #  - <workflow_type>_output_postfix
    #  - <workflow_type>_workflow_run_context
    #  - <workflow_type>_post_submit_delay
    #  - <workflow_type>_dump_intermediate_job_data
    #  - <workflow_type>_dump_intermediate_job_data
    #  - <workflow_type>_check_job_completeness
    #  - <workflow_type>_check_job_completeness_delay
    #  - <workflow_type>_poll_callback

    def process_resources(self):
        """
        Method used by luigi to define the resources required when running this task to include into
        scheduling rules when using the central scheduler.
        """
        if self.is_branch():
            return super(BaseRemoteWorkflow, self).process_resources()

        resources = self.workflow_proxy.process_resources()
        if self.include_member_resources:
            resources.update(self.resources)
        return resources

    def is_controlling_remote_jobs(self):
        """
        Returns *True* if the remote workflow is only controlling remote jobs instead of handling
        new ones. This is the case when either *cancel_jobs* or *cleanup_jobs* is *True*.
        """
        return self.cancel_jobs or self.cleanup_jobs

    def control_output_postfix(self):
        """
        Hook that should return a string that is inserted into the names of control output files.
        """
        return self.get_branches_repr()

    def create_job_dashboard(self):
        """
        Hook method to return a configured :py:class:`law.job.BaseJobDashboard` instance that will
        be used by the worflow.
        """
        return None

    def forward_dashboard_event(self, dashboard, job_data, event, job_num):
        """
        Hook to preprocess and publish dashboard events. By default, every event is passed to the
        dashboard's :py:meth:`law.job.dashboard.BaseJobDashboard.publish` method unchanged.
        """
        # possible events:
        #   - action.submit
        #   - action.cancel
        #   - status.pending
        #   - status.running
        #   - status.finished
        #   - status.retry
        #   - status.failed
        # forward to dashboard in any event by default
        return dashboard.publish(job_data, event, job_num)

    def modify_polling_status_line(self, status_line):
        """
        Hook to modify the status line that is printed during polling.
        """
        return status_line

    @property
    def accepts_messages(self):
        if not getattr(self, "workflow_proxy", None):
            return super(BaseRemoteWorkflow, self).accepts_messages

        return isinstance(self.workflow_proxy, BaseRemoteWorkflowProxy)

    def handle_scheduler_message(self, msg, _attr_value=None):
        """ handle_scheduler_message(msg)
        Hook that is called when a scheduler message *msg* is received. Returns *True* when the
        messages was handled, and *False* otherwise.

        Handled messages in addition to those defined in
        :py:meth:`law.workflow.base.BaseWorkflow.handle_scheduler_message`:

            - ``parallel_jobs = <int>``
            - ``walltime = <str/int/float>``
            - ``poll_fails = <int>``
            - ``poll_interval = <str/int/float>``
            - ``retries = <int>``
        """
        attr, value = _attr_value or (None, None)

        # handle "parallel_jobs"
        if attr is None:
            m = re.match(r"^\s*(parallel\_jobs)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "parallel_jobs"
                # the workflow proxy must be set here
                if not getattr(self, "workflow_proxy", None):
                    value = Exception("workflow_proxy not set yet")
                else:
                    try:
                        n = self.workflow_proxy._set_parallel_jobs(int(m.group(3)))
                        value = "unlimited" if n == self.workflow_proxy.n_parallel_max else str(n)
                    except ValueError as e:
                        value = e

        # handle "walltime"
        if attr is None:
            m = re.match(r"^\s*(walltime)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "walltime"
                try:
                    self.walltime = self.__class__.walltime.parse(m.group(3))
                    value = human_duration(hours=self.walltime, colon_format=True)
                except ValueError as e:
                    value = e

        # handle "poll_fails"
        if attr is None:
            m = re.match(r"^\s*(poll\_fails)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "poll_fails"
                try:
                    self.poll_fails = int(m.group(3))
                    value = self.poll_fails
                except ValueError as e:
                    value = e

        # handle "poll_interval"
        if attr is None:
            m = re.match(r"^\s*(poll\_interval)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "poll_interval"
                try:
                    self.poll_interval = self.__class__.poll_interval.parse(m.group(3))
                    value = human_duration(minutes=self.poll_interval, colon_format=True)
                except ValueError as e:
                    value = e

        # handle "retries"
        if attr is None:
            m = re.match(r"^\s*(retries)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "retries"
                try:
                    self.retries = int(m.group(3))
                    value = self.retries
                except ValueError as e:
                    value = e

        return super(BaseRemoteWorkflow, self).handle_scheduler_message(msg, (attr, value))



# coding: utf-8


import os
import sys
import re
import glob
import json
from setuptools import setup, find_packages


this_dir = os.path.dirname(os.path.abspath(__file__))


# package keyworkds
keywords = [
    "luigi", "workflow", "pipeline", "remote", "gfal", "submission", "cluster", "grid", "condor",
    "lsf", "glite", "arc", "sandboxing", "docker", "singularity",
]


# package classifiers
classifiers = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Development Status :: 4 - Beta",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: BSD License",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Information Technology",
    "Topic :: System :: Monitoring",
]


# helper to read non-empty, stripped lines from an opened file
def readlines(f):
    for line in f.readlines():
        if line.strip():
            yield line.strip()


# read the readme file
open_kwargs = {} if sys.version_info.major < 3 else {"encoding": "utf-8"}
with open(os.path.join(this_dir, "README.md"), "r", **open_kwargs) as f:
    long_description = f.read()


# load installation requirements
with open(os.path.join(this_dir, "requirements.txt"), "r") as f:
    install_requires = list(readlines(f))


# load package infos
pkg = {}
with open(os.path.join(this_dir, "law", "__version__.py"), "r") as f:
    exec(f.read(), pkg)


# install options
options = {}

# check for a custom executable for entry points
# note: when installing with pip, changing the executable in the shebang is only effective when
# running "pip install" with (e.g.) "--no-binary law" or "--no-binary :all:"
executable = os.getenv("LAW_INSTALL_EXECUTABLE", "")
if executable == "env":
    executable = "/usr/bin/env python"
elif re.match(r"^python(|\d|\d\.\d|\d\.\d\.\d)$", executable):
    executable = "/usr/bin/env " + executable
if executable:
    options["build_scripts"] = {"executable": executable}


# prepare console scripts from contrib packages
console_scripts = []
for contrib_init in glob.glob(os.path.join(this_dir, "law", "contrib", "*", "__init__.py")):
    # get the path of the scripts json file
    json_path = os.path.join(os.path.dirname(contrib_init), "console_scripts.json")
    if not os.path.exists(json_path):
        continue
    # read contents
    with open(json_path, "r") as f:
        _console_scripts = json.load(f)
    # store them
    for script_name, entry_point in _console_scripts.items():
        # adjust script name and entry point
        if not script_name.startswith("law_cms_"):
            script_name = "law_cms_" + script_name
        if not entry_point.startswith("law.contrib.cms."):
            entry_point = "law.contrib.cms." + entry_point
        # store
        console_scripts.append("{}={}".format(script_name, entry_point))


setup(
    name="law",
    version=pkg["__version__"],
    author=pkg["__author__"],
    author_email=pkg["__email__"],
    description=pkg["__doc__"].strip().split("\n")[0].strip(),
    license=pkg["__license__"],
    url=pkg["__contact__"],
    keywords=" ".join(keywords),
    classifiers=classifiers,
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4",
    zip_safe=False,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    scripts=["bin/law", "bin/law2", "bin/law3"],
    entry_points={"console_scripts": console_scripts},
    options=options,
)



# coding: utf-8

"""
"law config" cli subprogram.
"""


from law.config import Config
from law.util import abort


_cfg = Config.instance()


def setup_parser(sub_parsers):
    """
    Sets up the command line parser for the *config* subprogram and adds it to *sub_parsers*.
    """
    parser = sub_parsers.add_parser(
        "config",
        prog="law config",
        description="Configuration helper to get, set or remove a value from the law configuration "
        "file ({}).".format(_cfg.config_file),
    )

    parser.add_argument(
        "name",
        nargs="?",
        help="the name of the config in the format <section>[.<option>]; when empty, all section "
        "names are printed; when no option is given, all section options are printed",
    )
    parser.add_argument(
        "value",
        nargs="?",
        help="the value to set; when empty, the current value is printed",
    )
    parser.add_argument(
        "--remove",
        "-r",
        action="store_true",
        help="remove the config",
    )
    parser.add_argument(
        "--expand",
        "-e",
        action="store_true",
        help="expand variables when getting a value",
    )
    parser.add_argument(
        "--location",
        "-l",
        action="store_true",
        help="print the location of the configuration file and exit",
    )


def execute(args):
    """
    Executes the *config* subprogram with parsed commandline *args*.
    """
    cfg = Config.instance()

    # just print the file location?
    if args.location:
        print(cfg.config_file)
        return

    # print sections when none is given
    if not args.name:
        print("\n".join(cfg.sections()))
        return

    # print section options when none is given
    section, option = args.name.split(".", 1) if "." in args.name else (args.name, None)
    if not option:
        print("\n".join(cfg.options(section)))
        return

    # removal
    if args.remove:
        abort("config removal not yet implemented")

    # setting
    if args.value:
        abort("config setting not yet implemented")

    # getting
    print(cfg.get_default(section, option, expand_vars=args.expand, expand_user=args.expand))



# coding: utf-8
# flake8: noqa

__all__ = [
    "Register", "Task", "WrapperTask", "ExternalTask",
    "SandboxTask",
    "BaseWorkflow", "WorkflowParameter", "LocalWorkflow", "workflow_property",
    "dynamic_workflow_condition",
    "FileSystemTarget", "FileSystemFileTarget", "FileSystemDirectoryTarget",
    "LocalFileSystem", "LocalTarget", "LocalFileTarget", "LocalDirectoryTarget",
    "TargetCollection", "FileCollection", "SiblingFileCollection", "NestedSiblingFileCollection",
    "MirroredTarget", "MirroredFileTarget", "MirroredDirectoryTarget",
    "Sandbox", "BashSandbox", "VenvSandbox",
    "BaseJobManager", "BaseJobFileFactory", "JobInputFile", "JobArguments",
    "NO_STR", "NO_INT", "NO_FLOAT", "is_no_param", "get_param", "Parameter",
    "TaskInstanceParameter", "OptionalBoolParameter", "DurationParameter", "BytesParameter",
    "CSVParameter", "MultiCSVParameter", "RangeParameter", "MultiRangeParameter", "NotifyParameter",
    "NotifyMultiParameter", "NotifyMailParameter",
    "Config",
    "run", "no_value",
    "notify_mail",
    "luigi_version", "luigi_version_info",
]


import os

# package infos
from law.__version__ import (
    __doc__, __author__, __email__, __copyright__, __credits__, __contact__, __license__,
    __status__, __version__,
)

# luigi version infos
import re
import luigi

# __version__ was introduced in 2.8.11
luigi_version = getattr(luigi, "__version__", "2.8.10")
luigi_version_info = tuple(
    int(part) if i < 3 else part
    for i, part in enumerate(re.match(r"^(\d+)\.(\d+)\.(\d+)(.*)$", luigi_version).groups())
)


# setup logging
import law.logger
law.logger.setup_logging()


# prefer cached software
if os.getenv("LAW_USE_SOFTWARE_CACHE", "1").lower() in ("1", "yes", "true"):
    from law.cli.software import use_software_cache
    use_software_cache(reload_deps=True)


# luigi patches
import law.patches
law.patches.patch_all()


# provisioning imports
import law.util
from law.util import law_run as run, no_value
from law.config import Config
from law.notification import notify_mail
from law.parameter import (
    NO_STR, NO_INT, NO_FLOAT, is_no_param, get_param, Parameter, TaskInstanceParameter,
    OptionalBoolParameter, DurationParameter, BytesParameter, CSVParameter, MultiCSVParameter,
    RangeParameter, MultiRangeParameter, NotifyParameter, NotifyMultiParameter, NotifyMailParameter,
)
from law.target.file import (
    FileSystemTarget, FileSystemFileTarget, FileSystemDirectoryTarget, localize_file_targets,
)
from law.target.local import LocalFileSystem, LocalTarget, LocalFileTarget, LocalDirectoryTarget
from law.target.collection import (
    TargetCollection, FileCollection, SiblingFileCollection, NestedSiblingFileCollection,
)
from law.target.mirrored import MirroredTarget, MirroredFileTarget, MirroredDirectoryTarget
import law.decorator
from law.task.base import Register, Task, WrapperTask, ExternalTask
from law.workflow.base import (
    BaseWorkflow, WorkflowParameter, workflow_property, dynamic_workflow_condition,
)
from law.workflow.local import LocalWorkflow
from law.sandbox.base import Sandbox, SandboxTask
from law.sandbox.bash import BashSandbox
from law.sandbox.venv import VenvSandbox
from law.job.base import BaseJobManager, BaseJobFileFactory, JobInputFile, JobArguments
import law.job.dashboard
import law.workflow.remote
import law.contrib



# coding: utf-8

"""
CMS-related tasks.
https://home.cern/about/experiments/cms
"""

__all__ = ["BundleCMSSW"]


import os
import subprocess
from abc import abstractmethod

import luigi


from law.task.base import Task
from law.target.file import get_path
from law.target.local import LocalFileTarget
from law.parameter import NO_STR, CSVParameter
from law.decorator import log
from law.util import rel_path, interruptable_popen, quote_cmd


class BundleCMSSW(Task):

    task_namespace = "law.cms"

    exclude = luigi.Parameter(
        default=NO_STR,
        significant=False,
        description="regular expression for excluding files or directories relative to CMSSW_BASE; "
        "default: empty",
    )
    include = CSVParameter(
        default=(),
        significant=False,
        description="comma-separated list of files or directories relative to CMSSW_BASE to "
        "include; default: empty",
    )
    custom_checksum = luigi.Parameter(
        default=NO_STR,
        description="a custom checksum to use; default: empty",
    )

    cmssw_checksumming = True

    def __init__(self, *args, **kwargs):
        super(BundleCMSSW, self).__init__(*args, **kwargs)

        self._checksum = None

    @abstractmethod
    def get_cmssw_path(self):
        return

    @property
    def checksum(self):
        if not self.cmssw_checksumming:
            return None

        if self.custom_checksum != NO_STR:
            return self.custom_checksum

        if self._checksum is None:
            cmd = [
                rel_path(__file__, "scripts", "cmssw_checksum.sh"),
                get_path(self.get_cmssw_path()),
            ]
            if self.exclude != NO_STR:
                cmd += [self.exclude]
            cmd = quote_cmd(cmd)

            code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE)
            if code != 0:
                raise Exception("cmssw checksum calculation failed")

            self._checksum = out.strip()

        return self._checksum

    def output(self):
        base = os.path.basename(get_path(self.get_cmssw_path()))
        if self.checksum:
            base += "{}.".format(self.checksum)
        base = os.path.abspath(os.path.expandvars(os.path.expanduser(base)))
        return LocalFileTarget("{}.tgz".format(base))

    @log
    def run(self):
        with self.output().localize("w") as tmp:
            with self.publish_step("bundle CMSSW ..."):
                self.bundle(tmp.path)

    def get_cmssw_bundle_command(self, dst_path):
        return [
            rel_path(__file__, "scripts", "bundle_cmssw.sh"),
            get_path(self.get_cmssw_path()),
            get_path(dst_path),
            self.exclude if self.exclude not in (None, NO_STR) else "",
            " ".join(self.include),
        ]

    def bundle(self, dst_path):
        cmd = self.get_cmssw_bundle_command(dst_path)
        code = interruptable_popen(quote_cmd(cmd), shell=True, executable="/bin/bash")[0]
        if code != 0:
            raise Exception("cmssw bundling failed")



# coding: utf-8

"""
Crab job manager and CMS-related job helpers.
"""

__all__ = ["CrabJobManager", "CrabJobFileFactory", "CMSJobDashboard"]


import os
import stat
import time
import socket
import threading
import re
import json
import subprocess
import shutil
from collections import OrderedDict, namedtuple

import six

import law
from law.config import Config
from law.sandbox.base import Sandbox
from law.job.base import BaseJobManager, BaseJobFileFactory, JobInputFile, DeprecatedInputFiles
from law.job.dashboard import BaseJobDashboard
from law.target.file import get_path
from law.util import (
    DotDict, interruptable_popen, make_list, make_unique, quote_cmd, no_value, rel_path,
)
from law.logger import get_logger

import law.contrib.cms.sandbox


law.contrib.load("wlcg")

logger = get_logger(__name__)


class CrabJobManager(BaseJobManager):

    submission_task_name_cre = re.compile(r"^Task\s+name\s*\:\s+([^\s]+)\s*$")
    submission_log_file_cre = re.compile(r"^Log\s+file\s+is\s+([^\s]+\.log)\s*$")
    query_server_status_cre = re.compile(r"^Status\s+on\s+the\s+CRAB\s+server\s*\:\s+([^\s].*)$")
    query_user_cre = re.compile(r"^Task\s+name\s*:\s+\d+_\d+\:([^_]+)_.+$")
    query_scheduler_cre = re.compile(r"^Grid\s+scheduler\s+-\s+Task\s+Worker\s*\:\s+([^\s]+).+$")
    query_scheduler_id_cre = re.compile(r"^Grid\s+scheduler\s+-\s+Task\s+Worker\s*\:\s+crab.*\@.+[^\d](\d+)\..+$")  # noqa
    query_scheduler_status_cre = re.compile(r"^Status\s+on\s+the\s+scheduler\s*\:\s+([^\s].*)$")
    query_monitoring_url_cre = re.compile(r"^Dashboard\s+monitoring\s+URL\s*\:\s+([^\s].*)$")
    query_json_line_cre = re.compile(r"^\s*(\{.+\})\s*$")
    log_n_jobs_cre = re.compile(r"^config\.Data\.totalUnits\s+\=\s+(\d+)\s*$")
    log_disable_output_collection_cre = re.compile(r"^config\.JobType\.disableAutomaticOutputCollection\s+\=\s+([^\s]+)\s*$")  # noqa
    log_task_name_cre = re.compile(r"^.+\s+Task\s+name\s*\:\s+([^\s]+)\s*$")

    log_file_pattern = "https://cmsweb.cern.ch:8443/scheddmon/{scheduler_id}/{user}/{task_name}/job_out.{crab_num}.{attempt}.txt"  # noqa

    job_grouping_submit = True
    job_grouping_query = True
    job_grouping_cancel = True
    job_grouping_cleanup = True

    JobId = namedtuple("JobId", ["crab_num", "task_name", "proj_dir"])

    def __init__(self, sandbox_name=None, proxy=None, instance=None, threads=1):
        super(CrabJobManager, self).__init__()

        # default sandbox name
        if sandbox_name is None:
            cfg = Config.instance()
            sandbox_name = cfg.get_expanded("job", "crab_sandbox_name")

        # create the cmssw sandbox
        self.cmssw_sandbox = Sandbox.new(
            sandbox_name
            if sandbox_name.startswith("cmssw::")
            else "cmssw::{}".format(sandbox_name),
        )

        # store attributes
        self.proxy = proxy
        self.instance = instance
        self.threads = threads

    @classmethod
    def cast_job_id(cls, job_id):
        """
        Converts a *job_id*, for instance after json deserialization, into a :py:class:`JobId`
        object.
        """
        return cls.JobId(*job_id) if isinstance(job_id, (list, tuple)) else job_id

    @property
    def cmssw_env(self):
        return self.cmssw_sandbox.env

    def group_job_ids(self, job_ids):
        groups = OrderedDict()

        # group by project directory
        for job_id in job_ids:
            if job_id.proj_dir not in groups:
                groups[job_id.proj_dir] = []
            groups[job_id.proj_dir].append(job_id)

        return groups

    def _apply_group(self, func, result_type, group_func, job_objs, *args, **kwargs):
        # when job_objs is a string or a sequence of strings, interpret them as project dirs, read
        # their log files to extract task names, build actual job ids and forward them
        if func != self.submit:
            job_ids = []
            for i, job_id in enumerate(make_list(job_objs)):
                if not isinstance(job_id, six.string_types):
                    job_ids.append(job_id)
                    continue

                # get n_jobs and task_name from log file
                proj_dir = job_id
                log_file = os.path.join(proj_dir, "crab.log")
                log_data = self._parse_log_file(log_file)
                if not log_data or "n_jobs" not in log_data or "task_name" not in log_data:
                    job_ids.append(job_id)
                    continue

                # expand ids
                for crab_num in range(1, int(log_data["n_jobs"]) + 1):
                    job_ids.append(self.JobId(crab_num, log_data["task_name"], proj_dir))
            job_objs = job_ids

        return super(CrabJobManager, self)._apply_group(
            func,
            result_type,
            group_func,
            job_objs,
            *args,
            **kwargs  # noqa
        )

    def submit(self, job_file, job_files=None, proxy=None, instance=None, myproxy_username=None,
            retries=0, retry_delay=3, silent=False):
        # default arguments
        if proxy is None:
            proxy = self.proxy
        if instance is None:
            instance = self.instance

        # get the job file location as the submission command is run it the same directory
        job_file_dir, job_file_name = os.path.split(os.path.abspath(str(job_file)))

        # define the actual submission in a loop to simplify retries
        while True:
            # build the command
            cmd = ["crab", "submit", "--config", job_file_name]
            if proxy:
                cmd += ["--proxy", proxy]
            if instance:
                cmd += ["--instance", instance]
            cmd = quote_cmd(cmd)

            # run the command
            # crab prints everything to stdout
            logger.debug("submit crab jobs with command '{}'".format(cmd))
            code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE, cwd=job_file_dir, env=self.cmssw_env)

            # handle errors
            if code != 0:
                logger.debug("submission of crab job '{}' failed with code {}:\n{}".format(
                    job_file, code, out))

                # remove the project directory
                proj_dir = self._proj_dir_from_job_file(job_file, self.cmssw_env)
                if proj_dir and os.path.isdir(proj_dir):
                    logger.debug("removing crab project '{}' from previous attempt".format(
                        proj_dir))
                    shutil.rmtree(proj_dir)

                if retries > 0:
                    retries -= 1
                    time.sleep(retry_delay)
                    continue

                if silent:
                    return None

                raise Exception("submission of crab job '{}' failed with code {}:\n{}".format(
                    job_file, code, out))

            # parse outputs
            task_name, log_file = None, None
            for line in out.replace("\r", "").split("\n"):
                if not task_name:
                    m_task_name = self.submission_task_name_cre.match(line)
                    if m_task_name:
                        task_name = m_task_name.group(1)

                if not log_file:
                    m_log_file = self.submission_log_file_cre.match(line)
                    if m_log_file:
                        log_file = m_log_file.group(1)

                if task_name and log_file:
                    break

            if not task_name:
                raise Exception("no valid task name found in submission output:\n\n{}".format(out))
            if not log_file:
                raise Exception("no valid log file found in submission output:\n\n{}".format(out))

            # create job ids with log data
            proj_dir = os.path.dirname(log_file)
            job_ids = self._job_ids_from_proj_dir(proj_dir)

            # checks
            if not job_ids:
                raise Exception("number of jobs not extractable from log file {}".format(log_file))
            if job_files is not None and len(job_files) != len(job_ids):
                raise Exception(
                    "number of submited jobs ({}) does not match number of job files ({})".format(
                        len(job_ids), len(job_file)),
                )

            return job_ids

    def cancel(self, proj_dir, job_ids=None, proxy=None, instance=None, myproxy_username=None,
            silent=False):
        if job_ids is None:
            job_ids = self._job_ids_from_proj_dir(proj_dir)

        # build the command
        cmd = ["crab", "kill", "--dir", str(proj_dir)]
        if proxy:
            cmd += ["--proxy", proxy]
        if instance:
            cmd += ["--instance", instance]
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("cancel crab job(s) with command '{}'".format(cmd))
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, env=self.cmssw_env)

        # check success
        if code != 0 and not silent:
            # crab prints everything to stdout
            raise Exception(
                "cancellation of crab jobs from project '{}' failed with code {}:\n{}".format(
                    proj_dir, code, out),
            )

        return {job_id: None for job_id in job_ids}

    def cleanup(self, proj_dir, job_ids=None, proxy=None, instance=None, myproxy_username=None,
            silent=False):
        if job_ids is None:
            job_ids = self._job_ids_from_proj_dir(proj_dir)

        # just delete the project directory
        proj_dir = str(proj_dir)
        if os.path.isdir(proj_dir):
            shutil.rmtree(proj_dir)

        return {job_id: None for job_id in job_ids}

    def query(self, proj_dir, job_ids=None, proxy=None, instance=None, myproxy_username=None,
            skip_transfers=None, silent=False):
        proj_dir = str(proj_dir)
        log_data = self._parse_log_file(os.path.join(proj_dir, "crab.log"))
        if job_ids is None:
            job_ids = self._job_ids_from_proj_dir(proj_dir, log_data=log_data)

        # when output collection is disabled, we can consider all "transferring" states as finished
        if skip_transfers is None:
            skip_transfers = str(log_data.get("disable_output_collection")).lower() == "true"

        # build the command
        cmd = ["crab", "status", "--dir", proj_dir, "--json"]
        if proxy:
            cmd += ["--proxy", proxy]
        if instance:
            cmd += ["--instance", instance]
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("query crab job(s) with command '{}'".format(cmd))
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, env=self.cmssw_env)

        # handle errors
        if code != 0:
            if silent:
                return None
            # crab prints everything to stdout
            raise Exception(
                "status query of crab jobs from project '{}' failed with code {}:\n{}".format(
                    proj_dir, code, out),
            )

        # parse the output and extract the status per job
        query_data = self.parse_query_output(out, proj_dir, job_ids, skip_transfers=skip_transfers)

        # compare to the requested job ids and perform some checks
        for job_id in job_ids:
            if job_id not in query_data:
                query_data[job_id] = self.job_status_dict(
                    job_id=job_id,
                    status=self.FAILED,
                    error="job not found in query response",
                )

        return query_data

    @classmethod
    def parse_query_output(cls, out, proj_dir, job_ids, skip_transfers=False):
        # parse values using compiled regexps
        cres = [
            cls.query_user_cre,
            cls.query_server_status_cre,
            cls.query_scheduler_id_cre,
            cls.query_scheduler_status_cre,
            cls.query_json_line_cre,
            cls.query_monitoring_url_cre,
        ]
        values = len(cres) * [None]
        for line in out.replace("\r", "").split("\n"):
            for i, (cre, value) in enumerate(zip(cres, values)):
                if value:
                    continue
                m = cre.match(line)
                if m:
                    values[i] = m.group(1)
            if all(values):
                break

        # unpack
        username, server_status, scheduler_id, scheduler_status, json_line, monitoring_url = values

        # helper to build extra info
        def extra(job_id, job_data=None):
            extra = {}
            if username and scheduler_id and job_data:
                extra["log_file"] = cls.log_file_pattern.format(
                    scheduler_id=scheduler_id,
                    user=username,
                    task_name=job_id.task_name,
                    crab_num=job_id.crab_num,
                    attempt=job_data.get("Retries", 0),
                )
            if monitoring_url:
                extra["tracking_url"] = monitoring_url
            return extra

        # in case scheduler status or the json line is missing, the submission could be too new
        if not scheduler_status or not json_line:
            accepted_server_states = [
                "HOLDING on command SUBMIT",
                "NEW on command SUBMIT",
                "QUEUED on command SUBMIT",
                "SUBMITTED",
            ]
            if server_status not in accepted_server_states:
                s = ",".join(map("'{}'".format, accepted_server_states))
                raise Exception(
                    "no per-job information available (yet?), which is only accepted if the crab " +
                    "server status is any of {}, but got '{}'".format(s, server_status),
                )
            # interpret all jobs as pending
            return {
                job_id: cls.job_status_dict(job_id=job_id, status=cls.PENDING, extra=extra(job_id))
                for job_id in job_ids
            }

        # parse json data
        if not json_line:
            raise Exception(
                "no per-job information available in status response, crab server " +
                "status '{}', scheduler status '{}'".format(server_status, scheduler_status),
            )

        # map of crab job numbers to full ids for faster lookup
        num_to_id_map = {job_id.crab_num: job_id for job_id in job_ids}

        # build query data
        query_data = {}
        for crab_num_str, data in json.loads(json_line).items():
            crab_num = int(crab_num_str)
            if crab_num not in num_to_id_map:
                continue
            job_id = num_to_id_map[crab_num]

            # parse error info
            code = None
            error = data.get("Error")
            if isinstance(error, list) and len(error) >= 2:
                code = error[0]
                error = str(error[1]).strip()

            # extra info
            _extra = extra(job_id, data) or {}
            _extra["site_history"] = data.get("SiteHistory")

            # fill query data
            query_data[job_id] = cls.job_status_dict(
                job_id=job_id,
                status=cls.map_status(data["State"], skip_transfers=skip_transfers),
                code=code,
                error=error,
                extra=_extra,
            )

        return query_data

    @classmethod
    def _proj_dir_from_job_file(cls, job_file, cmssw_env):
        work_area = None
        request_name = None

        job_file = str(job_file)
        with open(job_file, "r") as f:
            # fast approach: parse the job file
            for line in f.readlines():
                if work_area and request_name:
                    break
                if not work_area:
                    m = re.match(r"^.+[^\w]workArea\s*=\s(\'|\")(.+)(\'|\").*$", line.strip())
                    if m:
                        work_area = m.group(2)
                        continue
                if not request_name:
                    m = re.match(r"^.+[^\w]requestName\s*=\s(\'|\")(.+)(\'|\").*$", line.strip())
                    if m:
                        request_name = m.group(2)
                        continue

            # when the combination is correct, return
            if work_area and request_name and 0:
                path = os.path.join(work_area, "crab_{}".format(request_name))
                path = os.path.expandvars(os.path.expanduser(path))
                if os.path.isdir(path):
                    return path

            # long approach: read the file in the cmssw env and manually print
            m = re.match(r"^CMSSW(_.+|)_(\d)+_\d+_\d+.*$", cmssw_env["CMSSW_VERSION"])
            cmssw_major = int(m.group(2)) if m else None
            py_exec = "python3" if cmssw_major is None or cmssw_major >= 11 else "python"
            cmd = """{} -c '
from os.path import join
with open("{}", "r") as f:
    mod = dict()
    exec(f.read(), mod)
    cfg = mod["cfg"]
print(join(cfg.General.workArea, "crab_" + cfg.General.requestName))'""".format(py_exec, job_file)
            code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE, env=cmssw_env)
            if code == 0:
                path = out.strip().replace("\r\n", "\n").split("\n")[-1]
                path = os.path.expandvars(os.path.expanduser(path))
                if os.path.isdir(path):
                    return path

        return None

    @classmethod
    def _parse_log_file(cls, log_file):
        log_file = os.path.expandvars(os.path.expanduser(str(log_file)))
        if not os.path.exists(log_file):
            return None

        cres = [cls.log_n_jobs_cre, cls.log_task_name_cre, cls.log_disable_output_collection_cre]
        names = ["n_jobs", "task_name", "disable_output_collection"]
        values = len(cres) * [None]

        with open(log_file, "r") as f:
            for line in f.readlines():
                for i, (cre, value) in enumerate(zip(cres, values)):
                    if value:
                        continue
                    m = cre.match(line)
                    if m:
                        values[i] = m.group(1)
                if all(values):
                    break

        return dict(zip(names, values))

    @classmethod
    def _job_ids_from_proj_dir(cls, proj_dir, log_data=None):
        # read log data
        proj_dir = str(proj_dir)
        if not log_data:
            log_data = cls._parse_log_file(os.path.join(proj_dir, "crab.log"))
        if not log_data or "n_jobs" not in log_data or "task_name" not in log_data:
            return None

        # build and return ids
        return [
            cls.JobId(crab_num, log_data["task_name"], proj_dir)
            for crab_num in range(1, int(log_data["n_jobs"]) + 1)
        ]

    @classmethod
    def map_status(cls, status, skip_transfers=False):
        # see https://twiki.cern.ch/twiki/bin/view/CMSPublic/Crab3HtcondorStates
        if status in ("cooloff", "unsubmitted", "idle"):
            return cls.PENDING
        if status in ("running",):
            return cls.RUNNING
        if status in ("transferring", "transferred"):
            return cls.FINISHED if skip_transfers else cls.RUNNING
        if status in ("finished",):
            return cls.FINISHED
        if status in ("killing", "failed", "held"):
            return cls.FAILED
        return cls.FAILED


class CrabJobFileFactory(BaseJobFileFactory):

    config_attrs = BaseJobFileFactory.config_attrs + [
        "file_name", "executable", "arguments", "work_area", "request_name", "input_files",
        "output_files", "storage_site", "output_lfn_base", "vo_group", "vo_role", "crab",
        "custom_content", "absolute_paths",
    ]

    def __init__(self, file_name="crab_job.py", executable=None, arguments=None, work_area=None,
            request_name=None, input_files=None, output_files=None, storage_site=None,
            output_lfn_base=None, vo_group=None, vo_role=None, custom_content=None,
            absolute_paths=False, **kwargs):
        # get some default kwargs from the config
        cfg = Config.instance()
        default_dir = cfg.get_expanded(
            "job",
            cfg.find_option("job", "crab_job_file_dir", "job_file_dir"),
        )
        if kwargs.get("dir") is None:
            kwargs["dir"] = cfg.get_expanded(
                "job",
                cfg.find_option("job", "crab_job_file_dir", "job_file_dir"),
            )
        if kwargs.get("cleanup") is None:
            kwargs["cleanup"] = cfg.get_expanded_bool(
                "job",
                cfg.find_option("job", "crab_job_file_dir_cleanup", "job_file_dir_cleanup"),
            )
        if kwargs.get("mkdtemp") is None:
            kwargs["mkdtemp"] = cfg.get_expanded_bool(
                "job",
                cfg.find_option("job", "crab_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"),
            )

        super(CrabJobFileFactory, self).__init__(**kwargs)

        self.file_name = file_name
        self.executable = executable
        self.arguments = arguments
        self.work_area = work_area
        self.request_name = request_name
        self.input_files = DeprecatedInputFiles(input_files or {})
        self.output_files = output_files or []
        self.storage_site = storage_site
        self.output_lfn_base = output_lfn_base
        self.vo_group = vo_group
        self.vo_role = vo_role
        self.custom_content = custom_content
        self.absolute_paths = absolute_paths

        # defaults
        if not self.work_area:
            self.work_area = default_dir

        # crab config
        # "no_value" marks required settings, None marks optional settings
        self.crab = DotDict([
            ("General", DotDict([
                ("requestName", no_value),
                ("workArea", no_value),
                ("transferLogs", False),
                ("transferOutputs", no_value),
            ])),
            ("JobType", DotDict([
                ("pluginName", "Analysis"),
                ("psetName", rel_path(__file__, "crab", "PSet.py")),
                ("scriptExe", no_value),
                ("maxMemoryMB", 2048),
                ("allowUndistributedCMSSW", True),
                ("sendPythonFolder", False),
                ("disableAutomaticOutputCollection", True),
                ("inputFiles", no_value),
                ("outputFiles", no_value),
            ])),
            ("Data", DotDict([
                ("inputDBS", "global"),
                ("splitting", "FileBased"),
                ("unitsPerJob", 1),
                ("totalUnits", no_value),
                ("inputDataset", None),
                ("userInputFiles", None),
                ("allowNonValidInputDataset", True),
                ("outLFNDirBase", no_value),
                ("publication", False),
                ("ignoreLocality", False),
            ])),
            ("Site", DotDict([
                ("storageSite", no_value),
            ])),
            ("User", DotDict([
                ("voGroup", None),
                ("voRole", None),
            ])),
        ])

    def create(self, **kwargs):
        # merge kwargs and instance attributes
        c = self.get_config(**kwargs)

        # some sanity checks
        if not c.file_name:
            raise ValueError("file_name must not be empty")
        if not c.executable:
            raise ValueError("executable must not be empty")
        if not c.request_name:
            raise ValueError("request_name must not be empty")
        if "." in c.request_name:
            raise ValueError("request_name should not contain '.', got {}".format(c.request_name))
        if len(c.request_name) > 100:
            raise ValueError(
                "request_name must be less then 100 characters long, got {}: {}".format(
                    len(c.request_name), c.request_name),
            )
        if not c.output_lfn_base:
            raise ValueError("output_lfn_base must not be empty")
        if not c.storage_site:
            raise ValueError("storage_site must not be empty")
        if not isinstance(c.arguments, (list, tuple)):
            raise ValueError("arguments must be a list, got '{}'".format(c.arguments))
        if "job_file" not in c.input_files:
            raise ValueError("an input file with key 'job_file' is required")

        # ensure that all log files are output files
        for attr in ["custom_log_file"]:
            if c[attr] and c[attr] not in c.output_files:
                c.output_files.append(c[attr])
        c.output_files = list(map(str, c.output_files))

        # ensure that all input files are JobInputFile's
        c.input_files = {
            key: JobInputFile(f)
            for key, f in c.input_files.items()
        }

        # ensure that the executable is an input file, remember the key to access it
        if c.executable:
            executable_keys = [
                k
                for k, v in c.input_files.items()
                if get_path(v) == get_path(c.executable)
            ]
            if executable_keys:
                executable_key = executable_keys[0]
            else:
                executable_key = "executable_file"
                c.input_files[executable_key] = JobInputFile(c.executable)

        # add the wrapper file to the inputs
        c.input_files["crab_wrapper"] = JobInputFile(
            path=rel_path(__file__, "crab", "crab_wrapper.sh"),
            copy=True,
            render=True,
            increment=True,
        )

        # prepare input files
        def prepare_input(f):
            # when not copied, just return the absolute, original path
            abs_path = os.path.abspath(f.path)
            if f.copy:
                # copy the file and apply other transformations
                abs_path = self.provide_input(
                    src=abs_path,
                    dir=c.dir,
                    skip_existing=f.share,
                    increment_existing=f.increment and not f.share,
                )
            return abs_path

        # absolute input paths
        for key, f in c.input_files.items():
            f.path_sub_abs = prepare_input(f)

        # input paths relative to the submission dir
        for key, f in c.input_files.items():
            f.path_sub_rel = (
                os.path.basename(f.path_sub_abs)
                if f.copy and not c.absolute_paths
                else f.path_sub_abs
            )

        # input paths as seen by the job, before and after potential job-side rendering
        for key, f in c.input_files.items():
            f.path_job_pre_render = os.path.basename(f.path_sub_abs)
            f.path_job_post_render = f.path_job_pre_render

        # update files in render variables with that after potential rendering
        c.render_variables.update({
            key: f.path_job_post_render
            for key, f in c.input_files.items()
        })

        # add space separated input files before potential rendering to render variables
        c.render_variables["input_files"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
        )

        # add space separated list of input files for rendering
        c.render_variables["input_files_render"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
            if f.render_job
        )

        # add the custom log file to render variables
        if c.custom_log_file:
            c.render_variables["log_file"] = c.custom_log_file

        # inject arguments into the crab wrapper via render variables
        c.render_variables["crab_job_arguments_map"] = ("\n" + 8 * " ").join(
            "['{}']=\"{}\"".format(i + 1, str(args))
            for i, args in enumerate(c.arguments)
        )

        # linearize render variables
        render_variables = self.linearize_render_variables(c.render_variables)

        # prepare the job file
        job_file = self.postfix_input_file(os.path.join(c.dir, str(c.file_name)))

        # render copied input files
        for key, f in c.input_files.items():
            if not f.copy or not f.render_local:
                continue
            self.render_file(
                f.path_sub_abs,
                f.path_sub_abs,
                render_variables,
            )

        # prepare the executable when given
        if c.executable:
            c.executable = get_path(c.input_files[executable_key].path_job_post_render)
            # make the file executable for the user and group
            path = os.path.join(c.dir, os.path.basename(c.executable))
            if os.path.exists(path):
                os.chmod(path, os.stat(path).st_mode | stat.S_IXUSR | stat.S_IXGRP)

        # resolve work_area relative to self.dir
        if c.work_area:
            work_area = os.path.expandvars(os.path.expanduser(str(c.work_area)))
            c.work_area = os.path.join(self.dir, work_area)
        else:
            c.work_area = self.dir

        # General
        c.crab.General.requestName = c.request_name
        c.crab.General.workArea = c.work_area
        c.crab.General.transferOutputs = bool(c.output_files)

        # JobType
        c.crab.JobType.scriptExe = c.input_files["crab_wrapper"].path_sub_rel
        c.crab.JobType.inputFiles = make_unique([
            f.path_sub_rel
            for f in c.input_files.values()
        ])
        c.crab.JobType.outputFiles = make_unique(c.output_files) or None

        # Data
        c.crab.Data.totalUnits = len(c.arguments)
        c.crab.Data.outLFNDirBase = c.output_lfn_base
        # define custom input files when no inputDataset is given
        # note: they don't have to exist but crab requires a list of length totalUnits
        if not c.crab.Data.inputDataset:
            c.crab.Data.userInputFiles = [
                "input_{}.root".format(i + 1)
                for i in range(c.crab.Data.totalUnits)
            ]

        # Site
        c.crab.Site.storageSite = c.storage_site

        # User
        if c.vo_group:
            c.crab.User.voGroup = c.vo_group
        if c.vo_role:
            c.crab.User.voRole = c.vo_role

        # write the job file
        self.write_crab_config(job_file, c.crab, custom_content=c.custom_content)

        logger.debug("created crab job file at '{}'".format(job_file))

        return job_file, c

    @classmethod
    def write_crab_config(cls, job_file, crab_config, custom_content=None):
        fmt_flat = lambda s: "\"{}\"".format(s) if isinstance(s, six.string_types) else str(s)

        with open(job_file, "w") as f:
            # header
            f.write("# coding: utf-8\n")
            f.write("\n")
            f.write("from CRABClient.UserUtilities import config\n")
            f.write("\n")
            f.write("cfg = config()\n")
            f.write("\n")

            # sections
            for section, cfg in crab_config.items():
                f.write("cfg.section_(\"{}\")\n".format(section))
                # options
                for option, value in cfg.items():
                    if value == no_value:
                        raise Exception(
                            "cannot assign {} to crab config {}.{}".format(value, section, option),
                        )
                    if value is None:
                        continue
                    value_str = (
                        "[\n{}\n]".format("\n".join("    {},".format(fmt_flat(v)) for v in value))
                        if isinstance(value, (list, tuple))
                        else fmt_flat(value)
                    )
                    f.write("cfg.{}.{} = {}\n".format(section, option, value_str))
                f.write("\n")

            # custom content
            if isinstance(custom_content, six.string_types):
                f.write(custom_content + "\n")
            elif isinstance(custom_content, (list, tuple)):
                for line in custom_content:
                    f.write(str(line) + "\n")


class CMSJobDashboard(BaseJobDashboard):
    """
    This CMS job dashboard interface requires ``apmon`` to be installed on your system.
    See http://monalisa.caltech.edu/monalisa__Documentation__ApMon_User_Guide__apmon_ug_py.html and
    https://twiki.cern.ch/twiki/bin/view/ArdaGrid/CMSJobMonitoringCollector.
    """

    PENDING = "pending"
    RUNNING = "running"
    CANCELLED = "cancelled"
    POSTPROC = "postproc"
    SUCCESS = "success"
    FAILED = "failed"

    tracking_url = "http://dashb-cms-job.cern.ch/dashboard/templates/task-analysis/#" + \
        "table=Jobs&p=1&activemenu=2&refresh=60&tid={dashboard_task_id}"

    persistent_attributes = ["task_id", "cms_user", "voms_user", "init_timestamp"]

    def __init__(self, task, cms_user, voms_user, apmon_config=None, log_level="WARNING",
            max_rate=20, task_type="analysis", site=None, executable="law", application=None,
            application_version=None, submission_tool="law", submission_type="direct",
            submission_ui=None, init_timestamp=None):
        super(CMSJobDashboard, self).__init__(max_rate=max_rate)

        # setup the apmon thread
        try:
            self.apmon = Apmon(apmon_config, self.max_rate, log_level)
        except ImportError as e:
            e.message += " (required for {})".format(self.__class__.__name__)
            e.args = (e.message,) + e.args[1:]
            raise e

        # get the task family for use as default application name
        task_family = task.get_task_family() if isinstance(task, law.Task) else task

        # mandatory (persistent) attributes
        self.task_id = task.task_id if isinstance(task, law.Task) else task
        self.cms_user = cms_user
        self.voms_user = voms_user
        self.init_timestamp = init_timestamp or self.create_timestamp()

        # optional attributes
        self.task_type = task_type
        self.site = site
        self.executable = executable
        self.application = application or task_family
        self.application_version = application_version or self.task_id.rsplit("_", 1)[1]
        self.submission_tool = submission_tool
        self.submission_type = submission_type
        self.submission_ui = submission_ui or socket.gethostname()

        # start the apmon thread
        self.apmon.daemon = True
        self.apmon.start()

    def __del__(self):
        if getattr(self, "apmon", None) and self.apmon.is_alive():
            self.apmon.stop()
            self.apmon.join()

    @classmethod
    def create_timestamp(cls):
        return time.strftime("%y%m%d_%H%M%S")

    @classmethod
    def create_dashboard_task_id(cls, task_id, cms_user, timestamp=None):
        if not timestamp:
            timestamp = cls.create_timestamp()
        return "{}:{}_{}".format(timestamp, cms_user, task_id)

    @classmethod
    def create_dashboard_job_id(cls, job_num, job_id, attempt=0):
        return "{}_{}_{}".format(job_num, job_id, attempt)

    @classmethod
    def params_from_status(cls, dashboard_status, fail_code=1):
        if dashboard_status == cls.PENDING:
            return {"StatusValue": "pending", "SyncCE": None}
        elif dashboard_status == cls.RUNNING:
            return {"StatusValue": "running"}
        elif dashboard_status == cls.CANCELLED:
            return {"StatusValue": "cancelled", "SyncCE": None}
        elif dashboard_status == cls.POSTPROC:
            return {"StatusValue": "running", "JobExitCode": 0}
        elif dashboard_status == cls.SUCCESS:
            return {"StatusValue": "success", "JobExitCode": 0}
        elif dashboard_status == cls.FAILED:
            return {"StatusValue": "failed", "JobExitCode": fail_code}
        else:
            raise ValueError("invalid dashboard status '{}'".format(dashboard_status))

    @classmethod
    def map_status(cls, job_status, event):
        # when starting with "status.", event must end with the job status
        if event.startswith("status.") and event.split(".", 1)[-1] != job_status:
            raise ValueError("event '{}' does not match job status '{}'".format(event, job_status))

        status = lambda attr: "status.{}".format(getattr(BaseJobManager, attr))

        return {
            "action.submit": cls.PENDING,
            "action.cancel": cls.CANCELLED,
            "custom.running": cls.RUNNING,
            "custom.postproc": cls.POSTPROC,
            "custom.failed": cls.FAILED,
            status("FINISHED"): cls.SUCCESS,
        }.get(event)

    def remote_hook_file(self):
        return law.util.rel_path(__file__, "scripts", "cmsdashb_hooks.sh")

    def remote_hook_data(self, job_num, attempt):
        data = [
            "task_id='{}'".format(self.task_id),
            "cms_user='{}'".format(self.cms_user),
            "voms_user='{}'".format(self.voms_user),
            "init_timestamp='{}'".format(self.init_timestamp),
            "job_num={}".format(job_num),
            "attempt={}".format(attempt),
        ]
        if self.site:
            data.append("site='{}'".format(self.site))
        return data

    def create_tracking_url(self):
        dashboard_task_id = self.create_dashboard_task_id(self.task_id, self.cms_user,
            self.init_timestamp)
        return self.tracking_url.format(dashboard_task_id=dashboard_task_id)

    def create_message(self, job_data, event, job_num, attempt=0, custom_params=None, **kwargs):
        # we need the voms user, which must start with "/CN="
        voms_user = self.voms_user
        if not voms_user:
            return
        if not voms_user.startswith("/CN="):
            voms_user = "/CN=" + voms_user

        # map to job status to a valid dashboard status
        dashboard_status = self.map_status(job_data.get("status"), event)
        if not dashboard_status:
            return

        # build the dashboard task id
        dashboard_task_id = self.create_dashboard_task_id(self.task_id, self.cms_user,
            self.init_timestamp)

        # build the id of the particular job
        dashboard_job_id = self.create_dashboard_job_id(job_num, job_data["job_id"],
            attempt=attempt)

        # build the parameters to send
        params = {
            "TaskId": dashboard_task_id,
            "JobId": dashboard_job_id,
            "GridJobId": job_data["job_id"],
            "CMSUser": self.cms_user,
            "GridName": voms_user,
            "JSToolUI": kwargs.get("submission_ui", self.submission_ui),
        }

        # add optional params
        params.update({
            "TaskType": kwargs.get("task_type", self.task_type),
            "SyncCE": kwargs.get("site", self.site),
            "Executable": kwargs.get("executable", self.executable),
            "Application": kwargs.get("application", self.application),
            "ApplicationVersion": kwargs.get("application_version", self.application_version),
            "JSTool": kwargs.get("submission_tool", self.submission_tool),
            "SubmissionType": kwargs.get("submission_type", self.submission_type),
        })

        # add status params
        params.update(self.params_from_status(dashboard_status, fail_code=job_data.get("code", 1)))

        # add custom params
        if custom_params:
            params.update(custom_params)

        # finally filter None's and convert everything to strings
        params = {key: str(value) for key, value in six.iteritems(params) if value is not None}

        return (dashboard_task_id, dashboard_job_id, params)

    @BaseJobDashboard.cache_by_status
    def publish(self, *args, **kwargs):
        message = self.create_message(*args, **kwargs)
        if message:
            self.apmon.send(*message)


apmon_lock = threading.Lock()


class Apmon(threading.Thread):

    default_config = {
        "cms-jobmon.cern.ch:8884": {
            "sys_monitoring": 0,
            "general_info": 0,
            "job_monitoring": 0,
        },
    }

    def __init__(self, config=None, max_rate=20, log_level="INFO"):
        super(Apmon, self).__init__()

        import apmon
        log_level = getattr(apmon.Logger, log_level.upper())
        self._apmon = apmon.ApMon(config or self.default_config, log_level)
        self._apmon.maxMsgRate = int(max_rate * 1.5)

        # hotfix of a bug occurring in apmon for too large pids
        for key, value in self._apmon.senderRef.items():
            value["INSTANCE_ID"] = value["INSTANCE_ID"] & 0x7fffffff

        self._max_rate = max_rate
        self._queue = six.moves.queue.Queue()
        self._stop_event = threading.Event()

    def send(self, *args, **kwargs):
        self._queue.put((args, kwargs))

    def _send(self, *args, **kwargs):
        self._apmon.sendParameters(*args, **kwargs)

    def stop(self):
        self._stop_event.set()

    def run(self):
        while True:
            # handling stopping
            self._stop_event.wait(0.5)
            if self._stop_event.is_set():
                break

            if self._queue.empty():
                continue

            with apmon_lock:
                while not self._queue.empty():
                    args, kwargs = self._queue.get()
                    self._send(*args, **kwargs)
                    time.sleep(1.0 / self._max_rate)



# coding: utf-8

"""
Custom luigi file system and target objects.
"""

__all__ = [
    "FileSystem", "FileSystemTarget", "FileSystemFileTarget", "FileSystemDirectoryTarget",
    "get_path", "get_scheme", "has_scheme", "add_scheme", "remove_scheme", "localize_file_targets",
]


import os
import sys
import re
from abc import abstractmethod, abstractproperty
from functools import partial
from contextlib import contextmanager

from law.config import Config
import law.target.luigi_shims as shims
from law.target.base import Target
from law.util import map_struct, create_random_string, human_bytes, no_value


class FileSystem(shims.FileSystem):

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        # reads a law config section and returns parsed file system configs
        cfg = Config.instance()

        if config is None:
            config = {}

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func):
            if option not in config or overwrite:
                config[option] = func(section, option)

        # read configs
        int_or_none = partial(cfg.get_expanded_int, default=None)
        add("has_permissions", cfg.get_expanded_bool)
        add("default_file_perm", int_or_none)
        add("default_dir_perm", int_or_none)
        add("create_file_dir", cfg.get_expanded_bool)

        return config

    def __init__(self, name=None, has_permissions=True, default_file_perm=None,
            default_dir_perm=None, create_file_dir=True, **kwargs):
        super(FileSystem, self).__init__(**kwargs)

        self.name = name
        self.has_permissions = has_permissions
        self.default_file_perm = default_file_perm
        self.default_dir_perm = default_dir_perm
        self.create_file_dir = create_file_dir

    def __repr__(self):
        return "{}(name={}, {})".format(self.__class__.__name__, self.name, hex(id(self)))

    def dirname(self, path):
        return os.path.dirname(str(path)) if path != "/" else None

    def basename(self, path):
        return os.path.basename(str(path)) if path != "/" else "/"

    def ext(self, path, n=1):
        # split the path
        parts = self.basename(path).lstrip(".").split(".")

        # empty extension in the trivial case or use the last n parts except for the first one
        return "" if len(parts) == 1 else ".".join(parts[1:][min(-n, 0):])

    def _unscheme(self, path):
        return remove_scheme(path)

    @abstractproperty
    def default_instance(self):
        return

    @abstractmethod
    def abspath(self, path):
        return

    @abstractmethod
    def stat(self, path, **kwargs):
        return

    @abstractmethod
    def exists(self, path, stat=False, **kwargs):
        return

    @abstractmethod
    def isdir(self, path, **kwargs):
        return

    @abstractmethod
    def isfile(self, path, **kwargs):
        return

    @abstractmethod
    def chmod(self, path, perm, silent=True, **kwargs):
        return

    @abstractmethod
    def remove(self, path, recursive=True, silent=True, **kwargs):
        return

    @abstractmethod
    def mkdir(self, path, perm=None, recursive=True, silent=True, **kwargs):
        return

    @abstractmethod
    def listdir(self, path, pattern=None, type=None, **kwargs):
        return

    @abstractmethod
    def walk(self, path, max_depth=-1, **kwargs):
        return

    @abstractmethod
    def glob(self, pattern, cwd=None, **kwargs):
        return

    @abstractmethod
    def copy(self, src, dst, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def move(self, src, dst, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    @contextmanager
    def open(self, path, mode, perm=None, dir_perm=None, **kwargs):
        return


class FileSystemTarget(Target, shims.FileSystemTarget):

    file_class = None
    directory_class = None

    def __init__(self, path, fs=None, **kwargs):
        if fs:
            self.fs = fs

        self._path = None
        self._unexpanded_path = None

        super(FileSystemTarget, self).__init__(path=path, **kwargs)

    def _repr_pairs(self, color=True):
        pairs = super(FileSystemTarget, self)._repr_pairs()

        # add the fs name
        if self.fs:
            pairs.append(("fs", self.fs.name))

        # add the path
        cfg = Config.instance()
        expand = cfg.get_expanded_bool("target", "expand_path_repr")
        pairs.append(("path", self.path if expand else self.unexpanded_path))

        # optionally add the file size
        if cfg.get_expanded_bool("target", "filesize_repr"):
            stat = self.exists(stat=True)
            pairs.append(("size", human_bytes(stat.st_size, fmt="{:.1f}{}") if stat else "-"))

        return pairs

    def _parent_args(self):
        return (), {}

    @property
    def unexpanded_path(self):
        return self._unexpanded_path

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        path = self.fs._unscheme(str(path))
        self._unexpanded_path = path
        self._path = os.path.expandvars(os.path.expanduser(self._unexpanded_path))

    @property
    def dirname(self):
        return self.fs.dirname(self.path)

    @property
    def absdirname(self):
        return self.fs.dirname(self.abspath)

    @property
    def basename(self):
        return self.fs.basename(self.path)

    @property
    def unique_basename(self):
        return "{}_{}".format(hex(self.hash)[2:], self.basename)

    @property
    def parent(self):
        # get the dirname, but favor the unexpanded one to propagate variables
        dirname = self.dirname
        unexpanded_dirname = self.fs.dirname(self.unexpanded_path)
        expanded_dirname = os.path.expandvars(os.path.expanduser(unexpanded_dirname))
        if unexpanded_dirname and self.fs.abspath(dirname) == self.fs.abspath(expanded_dirname):
            dirname = unexpanded_dirname

        args, kwargs = self._parent_args()
        return self.directory_class(dirname, *args, **kwargs) if dirname is not None else None

    def sibling(self, *args, **kwargs):
        parent = self.parent
        if not parent:
            raise Exception("cannot determine parent of {!r}".format(self))

        return parent.child(*args, **kwargs)

    def stat(self, **kwargs):
        return self.fs.stat(self.path, **kwargs)

    def exists(self, **kwargs):
        return self.fs.exists(self.path, **kwargs)

    def remove(self, silent=True, **kwargs):
        self.fs.remove(self.path, silent=silent, **kwargs)

    def chmod(self, perm, silent=False, **kwargs):
        self.fs.chmod(self.path, perm, silent=silent, **kwargs)

    def makedirs(self, *args, **kwargs):
        parent = self.parent
        return None if parent is None else parent.touch(*args, **kwargs)

    @abstractproperty
    def fs(self):
        return

    @abstractproperty
    def abspath(self):
        return

    @abstractmethod
    def uri(self, return_all=False, scheme=True, **kwargs):
        return

    @abstractmethod
    def touch(self, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def copy_to(self, dst, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def copy_from(self, src, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def move_to(self, dst, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def move_from(self, src, perm=None, dir_perm=None, **kwargs):
        return

    @abstractmethod
    def copy_to_local(self, *args, **kwargs):
        return

    @abstractmethod
    def copy_from_local(self, *args, **kwargs):
        return

    @abstractmethod
    def move_to_local(self, *args, **kwargs):
        return

    @abstractmethod
    def move_from_local(self, *args, **kwargs):
        return

    @abstractmethod
    @contextmanager
    def localize(self, mode="r", perm=None, dir_perm=None, tmp_dir=None, **kwargs):
        return

    @abstractmethod
    def load(self, *args, **kwargs):
        return

    @abstractmethod
    def dump(self, *args, **kwargs):
        return


class FileSystemFileTarget(FileSystemTarget):

    type = "f"

    def ext(self, n=1):
        return self.fs.ext(self.path, n=n)

    def open(self, mode, **kwargs):
        return self.fs.open(self.path, mode, **kwargs)

    def touch(self, **kwargs):
        # create the file via open without content
        with self.open("w", **kwargs) as f:
            f.write("")

    def copy_to(self, dst, perm=None, dir_perm=None, **kwargs):
        # TODO: complain when dst not local? forward to copy_from request depending on protocol?
        return self.fs.copy(self.path, get_path(dst), perm=perm, dir_perm=dir_perm, **kwargs)

    def copy_from(self, src, perm=None, dir_perm=None, **kwargs):
        if isinstance(src, FileSystemFileTarget):
            return src.copy_to(self.abspath, perm=perm, dir_perm=dir_perm, **kwargs)

        # when src is a plain string, let the fs handle it
        # TODO: complain when src not local? forward to copy_to request depending on protocol?
        return self.fs.copy(get_path(src), self.path, perm=perm, dir_perm=dir_perm, **kwargs)

    def move_to(self, dst, perm=None, dir_perm=None, **kwargs):
        # TODO: complain when dst not local? forward to copy_from request depending on protocol?
        return self.fs.move(self.path, get_path(dst), perm=perm, dir_perm=dir_perm, **kwargs)

    def move_from(self, src, perm=None, dir_perm=None, **kwargs):
        if isinstance(src, FileSystemFileTarget):
            return src.move_to(self.abspath, perm=perm, dir_perm=dir_perm, **kwargs)

        # when src is a plain string, let the fs handle it
        # TODO: complain when src not local? forward to copy_to request depending on protocol?
        return self.fs.move(get_path(src), self.path, perm=perm, dir_perm=dir_perm, **kwargs)


class FileSystemDirectoryTarget(FileSystemTarget):

    type = "d"

    open = None

    def _child_args(self, path, type):
        return (), {}

    def child(self, path, type=None, mktemp_pattern=False, **kwargs):
        if type not in (None, "f", "d"):
            raise ValueError("invalid child type, use 'f' or 'd'")

        # apply mktemp's feature to replace at least three consecutive 'X' with random characters
        path = get_path(path)
        if mktemp_pattern and "XXX" in path:
            repl = lambda m: create_random_string(l=len(m.group(1)))
            path = re.sub("(X{3,})", repl, path)

        unexpanded_path = os.path.join(self.unexpanded_path, path)
        path = os.path.join(self.path, path)
        if type == "f":
            cls = self.file_class
        elif type == "d":
            cls = self.__class__
        elif not self.fs.exists(path):
            raise Exception("cannot guess type of non-existing path '{}'".format(path))
        elif self.fs.isdir(path):
            cls = self.__class__
            type = "d"
        else:
            cls = self.file_class
            type = "f"

        args, _kwargs = self._child_args(path, type)
        _kwargs.update(kwargs)

        return cls(unexpanded_path, *args, **_kwargs)

    def listdir(self, **kwargs):
        return self.fs.listdir(self.path, **kwargs)

    def glob(self, pattern, **kwargs):
        return self.fs.glob(pattern, cwd=self.path, **kwargs)

    def walk(self, **kwargs):
        return self.fs.walk(self.path, **kwargs)

    def touch(self, **kwargs):
        kwargs.setdefault("silent", True)
        self.fs.mkdir(self.path, **kwargs)

    def copy_to(self, dst, perm=None, dir_perm=None, **kwargs):
        # create the target dir
        _dst = get_path(dst)
        if isinstance(dst, FileSystemDirectoryTarget):
            dst.touch(perm=dir_perm, **kwargs)
        else:
            # TODO: complain when dst not local? forward to copy_from request depending on protocol?
            self.fs.mkdir(_dst, perm=dir_perm, **kwargs)

        # walk and operate recursively
        for path, dirs, files, _ in self.walk(max_depth=0, **kwargs):
            # recurse through directories and files
            for basenames, type_flag in [(dirs, "d"), [files, "f"]]:
                for basename in basenames:
                    t = self.child(basename, type=type_flag)
                    t.copy_to(os.path.join(_dst, basename), perm=perm, dir_perm=dir_perm, **kwargs)

        return _dst

    def copy_from(self, src, perm=None, dir_perm=None, **kwargs):
        # when src is a directory target itself, forward to its copy_to implementation as it might
        # be more performant to use its own directory walking
        if isinstance(src, FileSystemDirectoryTarget):
            return src.copy_to(self, perm=perm, dir_perm=dir_perm, **kwargs)

        # create the target dir
        self.touch(perm=dir_perm, **kwargs)

        # when src is a plain string, let the fs handle it
        # walk and operate recursively
        # TODO: complain when src not local? forward to copy_from request depending on protocol?
        _src = get_path(src)
        for path, dirs, files, _ in self.fs.walk(_src, max_depth=0, **kwargs):
            # recurse through directories and files
            for basenames, type_flag in [(dirs, "d"), [files, "f"]]:
                for basename in basenames:
                    t = self.child(basename, type=type_flag)
                    t.copy_from(os.path.join(_src, basename), perm=perm, dir_perm=dir_perm, **kwargs)

        return self.abspath

    def move_to(self, dst, perm=None, dir_perm=None, **kwargs):
        # create the target dir
        _dst = get_path(dst)
        if isinstance(dst, FileSystemDirectoryTarget):
            dst.touch(perm=dir_perm, **kwargs)
        else:
            # TODO: complain when dst not local? forward to copy_from request depending on protocol?
            self.fs.mkdir(_dst, perm=dir_perm, **kwargs)

        # walk and operate recursively
        for path, dirs, files, _ in self.walk(max_depth=0, **kwargs):
            # recurse through directories and files
            for basenames, type_flag in [(dirs, "d"), [files, "f"]]:
                for basename in basenames:
                    t = self.child(basename, type=type_flag)
                    t.move_to(os.path.join(_dst, basename), perm=perm, dir_perm=dir_perm, **kwargs)

        # finally remove
        self.remove()

        return _dst

    def move_from(self, src, perm=None, dir_perm=None, **kwargs):
        # when src is a directory target itself, forward to its move_to implementation as it might
        # be more performant to use its own directory walking
        if isinstance(src, FileSystemDirectoryTarget):
            return src.move_to(self, perm=perm, dir_perm=dir_perm, **kwargs)

        # create the target dir
        self.touch(perm=dir_perm, **kwargs)

        # when src is a plain string, let the fs handle it
        # walk and operate recursively
        # TODO: complain when src not local? forward to copy_from request depending on protocol?
        _src = get_path(src)
        for path, dirs, files, _ in self.fs.walk(_src, max_depth=0, **kwargs):
            # recurse through directories and files
            for basenames, type_flag in [(dirs, "d"), [files, "f"]]:
                for basename in basenames:
                    t = self.child(basename, type=type_flag)
                    t.copy_from(os.path.join(_src, basename), perm=perm, dir_perm=dir_perm, **kwargs)

        # finally remove
        self.fs.remove(_src)

        return self.abspath


FileSystemTarget.file_class = FileSystemFileTarget
FileSystemTarget.directory_class = FileSystemDirectoryTarget


def get_path(target):
    if isinstance(target, FileSystemTarget):
        path = getattr(target, "abspath", no_value)
        if path != no_value:
            return path

    path = getattr(target, "path", no_value)
    if path != no_value:
        return path

    if target:
        return str(target)

    return target


def get_scheme(uri):
    # ftp://path/to/file -> ftp
    # /path/to/file -> None
    m = re.match(r"^(\w+)\:\/\/.*$", str(uri))
    return m.group(1) if m else None


def has_scheme(uri):
    return get_scheme(uri) is not None


def add_scheme(path, scheme):
    # adds a scheme to a path, if it does not already contain one
    path = str(path)
    return "{}://{}".format(scheme.rstrip(":/"), path) if not has_scheme(path) else path


def remove_scheme(uri):
    # ftp://path/to/file -> /path/to/file
    # /path/to/file -> /path/to/file
    return re.sub(r"^(\w+\:\/\/)", "", str(uri))


@contextmanager
def localize_file_targets(struct, *args, **kwargs):
    """
    Takes an arbitrary *struct* of targets, opens the contexts returned by their
    :py:meth:`FileSystemFileTarget.localize` implementations and yields their localized
    representations in the same structure as passed in *struct*. When the context is closed, the
    contexts of all localized targets are closed.
    """
    managers = []

    def enter(target):
        if callable(getattr(target, "localize", None)):
            manager = target.localize(*args, **kwargs)
            managers.append(manager)
            return manager.__enter__()

        return target

    # localize all targets, maintain the structure
    localized_targets = map_struct(enter, struct)

    # prepare exception info
    exc = None
    exc_info = (None, None, None)

    try:
        yield localized_targets

    except (Exception, KeyboardInterrupt) as e:
        exc = e
        exc_info = sys.exc_info()
        raise

    finally:
        exit_exc = []
        for manager in managers:
            try:
                manager.__exit__(*exc_info)
            except Exception as e:
                exit_exc.append(e)

        # when there was no exception during the actual yield and
        # an exception occured in one of the exit methods, raise the first one
        if not exc and exit_exc:
            raise exit_exc[0]



# coding: utf-8

"""
Shallow luigi target subclasses that consume arguments for use in multi-inheritance scenarios.
"""

import luigi


class Target(luigi.target.Target):

    def __init__(self, **kwargs):
        super(Target, self).__init__()


class FileSystem(luigi.target.FileSystem):

    def __init__(self, **kwargs):
        super(FileSystem, self).__init__()


class FileSystemTarget(luigi.target.FileSystemTarget, Target):

    def __init__(self, *args, **kwargs):
        path = str(args[0] if args else kwargs["path"])
        super(FileSystemTarget, self).__init__(path)


class LocalFileSystem(luigi.local_target.LocalFileSystem):

    def __init__(self, **kwargs):
        super(LocalFileSystem, self).__init__()


class LocalTarget(luigi.LocalTarget, FileSystemTarget):

    def __init__(self, **kwargs):
        super(LocalTarget, self).__init__(
            path=kwargs.get("path", None),
            format=kwargs.get("format", None),
            is_tmp=kwargs.get("is_tmp", False),
        )



#!/usr/bin/env python

"""
Script to trigger a myproxy delegation, e.g. for crab submission.
"""


def delegate(renew, endpoint, username, password_file, vo, voms_proxy):
    import law

    law.contrib.load("cms", "wlcg")

    # when not renewing, check if a previous delegation exists
    if not renew:
        info = law.wlcg.get_myproxy_info(
            endpoint=endpoint,
            username=username,
            silent=True,
        )
        if info:
            print("existing myproxy delegation found for username {}".format(info["username"]))
            return

    if voms_proxy and (renew or not law.wlcg.check_vomsproxy_validity()):
        print("\nrenewing voms-proxy")
        law.cms.renew_vomsproxy(vo=vo, password_file=password_file)

    # create a new delegation
    print("\ndelegating to {}".format(endpoint))
    law.cms.delegate_myproxy(
        endpoint=endpoint,
        username=username,
        password_file=password_file,
        vo=vo,
    )


def main():
    from argparse import ArgumentParser

    import law

    default_pf = law.config.get_expanded("job", "crab_password_file")

    parser = ArgumentParser(
        prog="law_cms_delegate_myproxy",
        description="delegates an X509 proxy to a myproxy server",
    )
    parser.add_argument(
        "--renew",
        "-r",
        action="store_true",
        help="renew the delegation even if an existing one was found",
    )
    parser.add_argument(
        "--endpoint",
        "-e",
        default="myproxy.cern.ch",
        help="the server endpoint; default: myproxy.cern.ch",
    )
    parser.add_argument(
        "--password-file",
        "-p",
        help="a file containing the certificate password" + (
            "; default: {}".format(default_pf) if default_pf else ""
        ),
        default=default_pf,
    )
    parser.add_argument(
        "--username",
        "-u",
        help="a custom user name",
    )
    parser.add_argument(
        "--vo",
        "-m",
        default="cms",
        help="virtual organization to use; default: cms",
    )
    parser.add_argument(
        "--voms-proxy",
        "-v",
        action="store_true",
        help="create a voms-proxy prior to the delegation",
    )
    args = parser.parse_args()

    delegate(
        args.renew,
        args.endpoint,
        args.username,
        args.password_file,
        args.vo,
        args.voms_proxy,
    )


# entry hook
if __name__ == "__main__":
    main()
